"""Variational quantum classifier.

In this demo we implement a variational classifier inspired by
Schuld et al. 2018 (arXiv:1804.00633).
"""

import pennylane as qml
from pennylane import numpy as np
from pennylane.optimize import NesterovMomentumOptimizer

dev = qml.device('default.qubit', wires=2)


def get_angles(x):
    """Compute rotation angles for state preparation routine.

    Args:
        x: original input vector

    Returns:
        array[float]: feature vector of angles
    """

    beta0 = 2 * np.arcsin(np.sqrt(x[1]) ** 2 / np.sqrt(x[0] ** 2 + x[1] ** 2 + 1e-12) )
    beta1 = 2 * np.arcsin(np.sqrt(x[3]) ** 2 / np.sqrt(x[2] ** 2 + x[3] ** 2 + 1e-12) )
    beta2 = 2 * np.arcsin(np.sqrt(x[2] ** 2 + x[3] ** 2) / np.sqrt(x[0] ** 2 + x[1] ** 2 + x[2] ** 2 + x[3] ** 2))

    return np.array([beta2, -beta1 / 2, beta1 / 2, -beta0 / 2, beta0 / 2])


def statepreparation(a):

    """Quantum circuit to encode a positive, normalized 4-d vector
    into the amplitudes of 2 qubits.

    Args:
        a: feature vector of rotation angles
    """

    # encode absolute values
    qml.RY(a[0], wires=0)

    qml.CNOT(wires=[0, 1])
    qml.RY(a[1], wires=1)
    qml.CNOT(wires=[0, 1])
    qml.RY(a[2], wires=1)

    qml.PauliX(wires=0)
    qml.CNOT(wires=[0, 1])
    qml.RY(a[3], wires=1)
    qml.CNOT(wires=[0, 1])
    qml.RY(a[4], wires=1)
    qml.PauliX(wires=0)


def layer(W):
    """ Single layer of the variational classifier.

    Args:
        W (array[float]): 2-d array of variables for one layer
    """

    qml.Rot(W[0, 0], W[0, 1], W[0, 2], wires=0)
    qml.Rot(W[1, 0], W[1, 1], W[1, 2], wires=1)

    qml.CNOT(wires=[0, 1])
    

@qml.qnode(dev)
def circuit(weights, angles=None):
    """The circuit of the variational classifier."""

    statepreparation(angles)
    
    for W in weights:
        layer(W)

    return qml.expval.PauliZ(0)


def variational_classifier(var, angles=None):
    """The variational classifier."""

    weights = var[0]
    bias = var[1]

    return circuit(weights, angles=angles) + bias


def square_loss(labels, predictions):
    """ Square loss function

    Args:
        labels (array[float]): 1-d array of labels
        predictions (array[float]): 1-d array of predictions
    Returns:
        float: square loss
    """
    loss = 0
    for l, p in zip(labels, predictions):
        loss = loss + (l - p) ** 2
    loss = loss / len(labels)

    return loss


def accuracy(labels, predictions):
    """ Share of equal labels and predictions

    Args:
        labels (array[float]): 1-d array of labels
        predictions (array[float]): 1-d array of predictions
    Returns:
        float: accuracy
    """

    loss = 0
    for l, p in zip(labels, predictions):
        if abs(l - p) < 1e-5:
            loss = loss + 1
    loss = loss / len(labels)

    return loss


def cost(weights, features, labels):
    """Cost (error) function to be minimized."""

    predictions = [variational_classifier(weights, angles=f) for f in features]

    return square_loss(labels, predictions)


# load Iris data and normalise feature vectors
data = np.loadtxt("data/iris_classes1and2_scaled.txt")
X = data[:, 0:2]

# pad the vectors to size 2^2 with constant values
padding = 0.3 * np.ones((len(X), 1))
X_pad = np.c_[np.c_[X, padding], np.zeros((len(X), 1)) ] 

# normalize each input
normalization = np.sqrt(np.sum(X_pad ** 2, -1))
X_norm = (X_pad.T / normalization).T  

# angles for state preparation are new features
features = np.array([get_angles(x) for x in X_norm])   

Y = data[:, -1]

# split into training and validation set
np.random.seed(0)
num_data = len(Y)
num_train = int(0.75 * num_data)
index = np.random.permutation(range(num_data))
feats_train = features[index[:num_train]]
Y_train = Y[index[:num_train]]
feats_val = features[index[num_train:]]
Y_val = Y[index[num_train:]]

# initialize weight layers
num_qubits = 2
num_layers = 6
var_init = (0.01 * np.random.randn(num_layers, num_qubits, 3), 0.0)

opt = NesterovMomentumOptimizer(0.01)
batch_size = 5

# train the variational classifier
var = var_init
for it in range(60):

    # Update the weights by one optimizer step
    batch_index = np.random.randint(0, num_train, (batch_size, ))
    feats_train_batch = feats_train[batch_index]
    Y_train_batch = Y_train[batch_index]
    var = opt.step(lambda v: cost(v, feats_train_batch, Y_train_batch), var)

    # Compute predictions on train and validation set
    predictions_train = [np.sign(variational_classifier(var, angles=f)) for f in feats_train]
    predictions_val = [np.sign(variational_classifier(var, angles=f)) for f in feats_val]

    # Compute accuracy on train and validation set
    acc_train = accuracy(Y_train, predictions_train)
    acc_val = accuracy(Y_val, predictions_val)

    print("Iter: {:5d} | Cost: {:0.7f} | Acc train: {:0.7f} | Acc validation: {:0.7f} "
          "".format(it+1, cost(var, features, Y), acc_train, acc_val))



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.plugin.DefaultQubit` device.
"""
# pylint: disable=protected-access,cell-var-from-loop
import unittest
import inspect
import logging as log

from pennylane import numpy as np

from defaults import pennylane as qml, BaseTest
from pennylane.plugins.default_qubit import (spectral_decomposition_qubit,
                                             I, X, Z, CNOT, Rphi, Rotx, Roty, Rotz, Rot3,
                                             unitary, hermitian, DefaultQubit)

log.getLogger('defaults')


U = np.array([[0.83645892-0.40533293j, -0.20215326+0.30850569j],
              [-0.23889780-0.28101519j, -0.88031770-0.29832709j]])


U2 = np.array([[-0.07843244-3.57825948e-01j, 0.71447295-5.38069384e-02j, 0.20949966+6.59100734e-05j, -0.50297381+2.35731613e-01j],
               [-0.26626692+4.53837083e-01j, 0.27771991-2.40717436e-01j, 0.41228017-1.30198687e-01j, 0.01384490-6.33200028e-01j],
               [-0.69254712-2.56963068e-02j, -0.15484858+6.57298384e-02j, -0.53082141+7.18073414e-02j, -0.41060450-1.89462315e-01j],
               [-0.09686189-3.15085273e-01j, -0.53241387-1.99491763e-01j, 0.56928622+3.97704398e-01j, -0.28671074-6.01574497e-02j]])


H = np.array([[1.02789352, 1.61296440-0.3498192j],
              [1.61296440+0.3498192j, 1.23920938+0j]])


def prep_par(par, op):
    "Convert par into a list of parameters that op expects."
    if op.par_domain == 'A':
        return [np.diag([x, 1]) for x in par]
    return par


class TestAuxillaryFunctions(BaseTest):
    """Test auxillary functions."""

    def test_spectral_decomposition_qubit(self):
        """Test that the correct spectral decomposition is returned."""
        self.logTestName()

        a, P = spectral_decomposition_qubit(H)

        # verify that H = \sum_k a_k P_k
        self.assertAllAlmostEqual(H, np.einsum('i,ijk->jk', a, P), delta=self.tol)

    def test_phase_shift(self):
        """Test phase shift is correct"""
        self.logTestName()

        # test identity for theta=0
        self.assertAllAlmostEqual(Rphi(0), np.identity(2), delta=self.tol)

        # test arbitrary phase shift
        phi = 0.5432
        expected = np.array([[1, 0], [0, np.exp(1j*phi)]])
        self.assertAllAlmostEqual(Rphi(phi), expected, delta=self.tol)

    def test_x_rotation(self):
        """Test x rotation is correct"""
        self.logTestName()

        # test identity for theta=0
        self.assertAllAlmostEqual(Rotx(0), np.identity(2), delta=self.tol)

        # test identity for theta=pi/2
        expected = np.array([[1, -1j], [-1j, 1]])/np.sqrt(2)
        self.assertAllAlmostEqual(Rotx(np.pi / 2), expected, delta=self.tol)

        # test identity for theta=pi
        expected = -1j*np.array([[0, 1], [1, 0]])
        self.assertAllAlmostEqual(Rotx(np.pi), expected, delta=self.tol)

    def test_y_rotation(self):
        """Test y rotation is correct"""
        self.logTestName()

        # test identity for theta=0
        self.assertAllAlmostEqual(Roty(0), np.identity(2), delta=self.tol)

        # test identity for theta=pi/2
        expected = np.array([[1, -1], [1, 1]])/np.sqrt(2)
        self.assertAllAlmostEqual(Roty(np.pi / 2), expected, delta=self.tol)

        # test identity for theta=pi
        expected = np.array([[0, -1], [1, 0]])
        self.assertAllAlmostEqual(Roty(np.pi), expected, delta=self.tol)

    def test_z_rotation(self):
        """Test z rotation is correct"""
        self.logTestName()

        # test identity for theta=0
        self.assertAllAlmostEqual(Rotz(0), np.identity(2), delta=self.tol)

        # test identity for theta=pi/2
        expected = np.diag(np.exp([-1j*np.pi/4, 1j*np.pi/4]))
        self.assertAllAlmostEqual(Rotz(np.pi / 2), expected, delta=self.tol)

        # test identity for theta=pi
        self.assertAllAlmostEqual(Rotz(np.pi), -1j * Z, delta=self.tol)

    def test_arbitrary_rotation(self):
        """Test arbitrary single qubit rotation is correct"""
        self.logTestName()

        # test identity for theta=0
        self.assertAllAlmostEqual(Rot3(0, 0, 0), np.identity(2), delta=self.tol)

        # expected result
        def arbitrary_rotation(x, y, z):
            """arbitrary single qubit rotation"""
            c = np.cos(y/2)
            s = np.sin(y/2)
            return np.array([[np.exp(-0.5j*(x+z))*c, -np.exp(0.5j*(x-z))*s],
                             [np.exp(-0.5j*(x-z))*s, np.exp(0.5j*(x+z))*c]])

        a, b, c = 0.432, -0.152, 0.9234
        self.assertAllAlmostEqual(Rot3(a, b, c), arbitrary_rotation(a, b, c), delta=self.tol)


class TestStateFunctions(BaseTest):
    """Arbitrary state and operator tests."""

    def test_unitary(self):
        """Test that the unitary function produces the correct output."""
        self.logTestName()

        out = unitary(U)

        # verify output type
        self.assertTrue(isinstance(out, np.ndarray))

        # verify equivalent to input state
        self.assertAllAlmostEqual(out, U, delta=self.tol)

        # test non-square matrix
        with self.assertRaisesRegex(ValueError, "must be a square matrix"):
            unitary(U[1:])

        # test non-unitary matrix
        U3 = U.copy()
        U3[0, 0] += 0.5
        with self.assertRaisesRegex(ValueError, "must be unitary"):
            unitary(U3)

    def test_hermitian(self):
        """Test that the Hermitian function produces the correct output."""
        self.logTestName()

        out = hermitian(H)

        # verify output type
        self.assertTrue(isinstance(out, np.ndarray))

        # verify equivalent to input state
        self.assertAllAlmostEqual(out, H, delta=self.tol)

        # test non-square matrix
        with self.assertRaisesRegex(ValueError, "must be a square matrix"):
            hermitian(H[1:])

        # test non-Hermitian matrix
        H2 = H.copy()
        H2[0, 1] = H2[0, 1].conj()
        with self.assertRaisesRegex(ValueError, "must be Hermitian"):
            hermitian(H2)


class TestDefaultQubitDevice(BaseTest):
    """Test the default qubit device. The test ensures that the device is properly
    applying qubit operations and calculating the correct observables."""
    def setUp(self):
        self.dev = DefaultQubit(wires=2, shots=0)

    def test_operation_map(self):
        """Test that default qubit device supports all PennyLane discrete gates."""
        self.logTestName()

        self.assertEqual(set(qml.ops.qubit.__all__), set(self.dev._operation_map))

    def test_expectation_map(self):
        """Test that default qubit device supports all PennyLane discrete expectations."""
        self.logTestName()

        self.assertEqual(set(qml.expval.qubit.__all__)|{'Identity'}, set(self.dev._expectation_map))

    def test_expand_one(self):
        """Test that a 1 qubit gate correctly expands to 3 qubits."""
        self.logTestName()

        dev = DefaultQubit(wires=3)

        # test applied to wire 0
        res = dev.expand_one(U, [0])
        expected = np.kron(np.kron(U, I), I)
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        # test applied to wire 1
        res = dev.expand_one(U, [1])
        expected = np.kron(np.kron(I, U), I)
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        # test applied to wire 2
        res = dev.expand_one(U, [2])
        expected = np.kron(np.kron(I, I), U)
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        # test exception raised if U is not 2x2 matrix
        with self.assertRaisesRegex(ValueError, "2x2 matrix required"):
            dev.expand_one(U2, [0])

        # test exception raised if more than one subsystem provided
        with self.assertRaisesRegex(ValueError, "One target subsystem required"):
            dev.expand_one(U, [0, 1])

    def test_expand_two(self):
        """Test that a 2 qubit gate correctly expands to 3 qubits."""
        self.logTestName()

        dev = DefaultQubit(wires=4)

        # test applied to wire 0+1
        res = dev.expand_two(U2, [0, 1])
        expected = np.kron(np.kron(U2, I), I)
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        # test applied to wire 1+2
        res = dev.expand_two(U2, [1, 2])
        expected = np.kron(np.kron(I, U2), I)
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        # test applied to wire 2+3
        res = dev.expand_two(U2, [2, 3])
        expected = np.kron(np.kron(I, I), U2)
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        # CNOT with target on wire 1
        res = dev.expand_two(CNOT, [1, 0])
        rows = np.array([0, 2, 1, 3])
        expected = np.kron(np.kron(CNOT[:, rows][rows], I), I)
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        # test exception raised if U is not 4x4 matrix
        with self.assertRaisesRegex(ValueError, "4x4 matrix required"):
            dev.expand_two(U, [0])

        # test exception raised if two subsystems are not provided
        with self.assertRaisesRegex(ValueError, "Two target subsystems required"):
            dev.expand_two(U2, [0])

        # test exception raised if unphysical subsystems provided
        with self.assertRaisesRegex(ValueError, "Bad target subsystems."):
            dev.expand_two(U2, [-1, 5])

    def test_get_operator_matrix(self):
        """Test the the correct matrix is returned given an operation name"""
        self.logTestName()

        for name, fn in {**self.dev._operation_map, **self.dev._expectation_map}.items():
            try:
                op = qml.ops.__getattribute__(name)
            except AttributeError:
                op = qml.expval.__getattribute__(name)

            p = [0.432423, -0.12312, 0.324][:op.num_params]

            if name == 'QubitStateVector':
                p = [np.array([1, 0, 0, 0])]
            elif name == 'QubitUnitary':
                p = [U]
            elif name == 'Hermitian':
                p = [H]

            res = self.dev._get_operator_matrix(name, p)

            if callable(fn):
                # if the default.qubit is an operation accepting parameters,
                # initialise it using the parameters generated above.
                expected = fn(*p)
            else:
                # otherwise, the operation is simply an array.
                expected = fn

            self.assertAllAlmostEqual(res, expected, delta=self.tol)

    def test_apply(self):
        """Test the application of gates to a state"""
        self.logTestName()
        self.dev.reset()

        # loop through all supported operations
        for gate_name, fn in self.dev._operation_map.items():
            log.debug("\tTesting %s gate...", gate_name)

            # start in the state |00>
            self.dev._state = np.array([1, 0, 0, 0])

            # get the equivalent pennylane operation class
            op = qml.ops.__getattribute__(gate_name)
            # the list of wires to apply the operation to
            w = list(range(op.num_wires))

            if op.par_domain == 'A':
                # the parameter is an array
                if gate_name == 'QubitStateVector':
                    p = [np.array([1, 0, 1, 1])/np.sqrt(3)]
                    expected_out = p
                elif gate_name == 'QubitUnitary':
                    p = [U]
                    w = [0]
                    expected_out = np.kron(U @ np.array([1, 0]), np.array([1, 0]))
                elif gate_name == 'BasisState':
                    p = [np.array([1, 1])]
                    expected_out = np.array([0, 0, 0, 1])

            elif op.par_domain == 'N':
                # the parameter is an integer
                p = [1, 3, 4][:op.num_params]
            else:
                # the parameter is a float
                p = [0.432423, -0.12312, 0.324][:op.num_params]

                if callable(fn):
                    # if the default.qubit is an operation accepting parameters,
                    # initialise it using the parameters generated above.
                    O = fn(*p)
                else:
                    # otherwise, the operation is simply an array.
                    O = fn

                # calculate the expected output
                if op.num_wires == 1:
                    expected_out = np.kron(O @ np.array([1, 0]), np.array([1, 0]))
                elif op.num_wires == 2:
                    expected_out = O @ self.dev._state

            self.dev.apply(gate_name, wires=w, par=p)

            # verify the device is now in the expected state
            self.assertAllAlmostEqual(self.dev._state, expected_out, delta=self.tol)

    def test_apply_errors(self):
        """Test that apply fails for incorrect state preparation, and > 2 qubit gates"""
        self.logTestName()

        with self.assertRaisesRegex(ValueError, r'State vector must be of length 2\*\*wires.'):
            p = [np.array([1, 0, 1, 1, 1])/np.sqrt(3)]
            self.dev.apply('QubitStateVector', wires=[0, 1], par=[p])

        with self.assertRaisesRegex(ValueError, "BasisState parameter must be an array of 0 or 1 integers of length at most 2."):
            self.dev.apply('BasisState', wires=[0, 1], par=[np.array([-0.2, 4.2])])

        with self.assertRaisesRegex(ValueError, "The default.qubit plugin can apply BasisState only to all of the 2 wires."):
            self.dev.apply('BasisState', wires=[0, 1, 2], par=[np.array([0, 1])])


        with self.assertRaisesRegex(ValueError, "This plugin supports only one- and two-qubit gates."):
            self.dev.apply('QubitUnitary', wires=[0, 1, 2], par=[U2])

    def test_ev(self):
        """Test that expectation values are calculated correctly"""
        self.logTestName()
        self.dev.reset()

        # loop through all supported observables
        for name, fn in self.dev._expectation_map.items():
            log.debug("\tTesting %s observable...", name)

            # start in the state |00>
            self.dev._state = np.array([1, 0, 1, 1])/np.sqrt(3)

            # get the equivalent pennylane operation class
            op = qml.expval.__getattribute__(name)

            if op.par_domain == 'A':
                # the parameter is an array
                p = [H]
            else:
                # the parameter is a float
                p = [0.432423, -0.12312, 0.324][:op.num_params]

            if callable(fn):
                # if the default.qubit is an operation accepting parameters,
                # initialise it using the parameters generated above.
                O = fn(*p)
            else:
                # otherwise, the operation is simply an array.
                O = fn

            print("op.num_wires="+str(op.num_wires))

            # calculate the expected output
            if op.num_wires == 1 or op.num_wires == 0:
                expected_out = self.dev._state.conj() @ np.kron(O, I) @ self.dev._state
            elif op.num_wires == 2:
                expected_out = self.dev._state.conj() @ O @ self.dev._state
            else:
                raise NotImplementedError("Test for operations with num_wires="+op.num_wires+" not implemented.")

            res = self.dev.ev(O, wires=[0])

            # verify the device is now in the expected state
            self.assertAllAlmostEqual(res, expected_out, delta=self.tol)

            # text exception raised if matrix is not 2x2
            with self.assertRaisesRegex(ValueError, "2x2 matrix required"):
                self.dev.ev(U2, [0])

            # text warning raised if matrix is complex
            with self.assertLogs(level='WARNING') as l:
                self.dev.ev(H+1j, [0])
                self.assertEqual(len(l.output), 1)
                self.assertEqual(len(l.records), 1)
                self.assertIn('Nonvanishing imaginary part', l.output[0])


class TestDefaultQubitIntegration(BaseTest):
    """Integration tests for default.qubit. This test ensures it integrates
    properly with the PennyLane interface, in particular QNode."""

    def test_load_default_qubit_device(self):
        """Test that the default plugin loads correctly"""
        self.logTestName()

        dev = qml.device('default.qubit', wires=2)
        self.assertEqual(dev.num_wires, 2)
        self.assertEqual(dev.shots, 0)
        self.assertEqual(dev.short_name, 'default.qubit')

    def test_args(self):
        """Test that the plugin requires correct arguments"""
        self.logTestName()

        with self.assertRaisesRegex(TypeError, "missing 1 required positional argument: 'wires'"):
            qml.device('default.qubit')

    def test_unsupported_gates(self):
        """Test error is raised with unsupported gates"""
        self.logTestName()
        dev = qml.device('default.qubit', wires=2)

        gates = set(dev._operation_map.keys())
        all_gates = {m[0] for m in inspect.getmembers(qml.ops, inspect.isclass)}

        for g in all_gates - gates:
            op = getattr(qml.ops, g)

            if op.num_wires == 0:
                wires = [0]
            else:
                wires = list(range(op.num_wires))

            @qml.qnode(dev)
            def circuit(*x):
                """Test quantum function"""
                x = prep_par(x, op)
                op(*x, wires=wires)

                if issubclass(op, qml.operation.CV):
                    return qml.expval.X(0)

                return qml.expval.PauliZ(0)

            with self.assertRaisesRegex(qml.DeviceError, "Gate {} not supported on device default.qubit".format(g)):
                x = np.random.random([op.num_params])
                circuit(*x)

    def test_unsupported_observables(self):
        """Test error is raised with unsupported observables"""
        self.logTestName()
        dev = qml.device('default.qubit', wires=2)

        obs = set(dev._expectation_map.keys())
        all_obs = set(qml.expval.__all__)

        for g in all_obs - obs:
            op = getattr(qml.expval, g)

            if op.num_wires == 0:
                wires = [0]
            else:
                wires = list(range(op.num_wires))

            @qml.qnode(dev)
            def circuit(*x):
                """Test quantum function"""
                x = prep_par(x, op)
                return op(*x, wires=wires)

            with self.assertRaisesRegex(qml.DeviceError, "Expectation {} not supported on device default.qubit".format(g)):
                x = np.random.random([op.num_params])
                circuit(*x)

    def test_qubit_circuit(self):
        """Test that the default qubit plugin provides correct result for simple circuit"""
        self.logTestName()
        dev = qml.device('default.qubit', wires=1)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.RX(x, wires=0)
            return qml.expval.PauliY(0)

        # <0|RX(p)^\dagger.PauliY.RX(p)|0> = -sin(p)
        expected = -np.sin(p)
        self.assertAlmostEqual(circuit(p), expected, delta=self.tol)

    def test_qubit_identity(self):
        """Test that the default qubit plugin provides correct result for the Identiy expectation"""
        self.logTestName()
        dev = qml.device('default.qubit', wires=1)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.RX(x, wires=0)
            return qml.expval.Identity(0)

        self.assertAlmostEqual(circuit(p), 1, delta=self.tol)

    def test_nonzero_shots(self):
        """Test that the default qubit plugin provides correct result for high shot number"""
        self.logTestName()

        shots = 10**4
        dev = qml.device('default.qubit', wires=1, shots=shots)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.RX(x, wires=0)
            return qml.expval.PauliY(0)

        runs = []
        for _ in range(100):
            runs.append(circuit(p))

        self.assertAlmostEqual(np.mean(runs), -np.sin(p), delta=0.01)

    def test_supported_gates(self):
        """Test that all supported gates work correctly"""
        self.logTestName()
        a = 0.312
        b = 0.123

        dev = qml.device('default.qubit', wires=2)

        for g, qop in dev._operation_map.items():
            log.debug('\tTesting gate %s...', g)
            self.assertTrue(dev.supported(g))
            dev.reset()

            op = getattr(qml.ops, g)
            if op.num_wires == 0:
                if g == 'BasisState':
                    wires = [0,1]
                else:
                    wires = [0]
            else:
                wires = list(range(op.num_wires))

            @qml.qnode(dev)
            def circuit(*x):
                """Reference quantum function"""
                op(*x, wires=wires)
                return qml.expval.PauliX(0)

            # compare to reference result
            def reference(*x):
                """reference circuit"""
                if callable(qop):
                    # if the default.qubit is an operation accepting parameters,
                    # initialise it using the parameters generated above.
                    O = qop(*x)
                else:
                    # otherwise, the operation is simply an array.
                    O = qop

                # calculate the expected output
                if op.num_wires == 1 or g == 'QubitUnitary':
                    out_state = np.kron(O @ np.array([1, 0]), np.array([1, 0]))
                elif g == 'QubitStateVector':
                    out_state = x[0]
                elif g == 'BasisState':
                    out_state = np.array([0, 0, 0, 1])
                else:
                    out_state = O @ dev._state

                expectation = out_state.conj() @ np.kron(X, np.identity(2)) @ out_state
                return expectation

            if g == 'QubitStateVector':
                p = np.array([1, 0, 0, 1])/np.sqrt(2)
                self.assertAllEqual(circuit(p), reference(p))
            elif g == 'BasisState':
                p = np.array([1, 1])
                self.assertAllEqual(circuit(p), reference(p))
            elif g == 'QubitUnitary':
                self.assertAllEqual(circuit(U), reference(U))
            elif op.num_params == 1:
                self.assertAllEqual(circuit(a), reference(a))
            elif op.num_params == 2:
                self.assertAllEqual(circuit(a, b), reference(a, b))

    def test_supported_observables(self):
        """Test that all supported observables work correctly"""
        self.logTestName()
        a = 0.312

        dev = qml.device('default.qubit', wires=2)

        for g, qop in dev._expectation_map.items():
            log.debug('\tTesting observable %s...', g)
            self.assertTrue(dev.supported(g))
            dev.reset()

            op = getattr(qml.expval, g)
            if op.num_wires == 0:
                wires = [0]
            else:
                wires = list(range(op.num_wires))

            @qml.qnode(dev)
            def circuit(*x):
                """Reference quantum function"""
                qml.RX(a, wires=0)
                return op(*x, wires=wires)

            # compare to reference result
            def reference(*x):
                """reference circuit"""
                if callable(qop):
                    # if the default.qubit is an operation accepting parameters,
                    # initialise it using the parameters generated above.
                    O = qop(*x)
                else:
                    # otherwise, the operation is simply an array.
                    O = qop

                # calculate the expected output
                out_state = np.kron(Rotx(a) @ np.array([1, 0]), np.array([1, 0]))
                expectation = out_state.conj() @ np.kron(O, np.identity(2)) @ out_state
                return expectation

            if op.num_params == 0:
                self.assertAllEqual(circuit(), reference())
            elif g == 'Hermitian':
                self.assertAllEqual(circuit(H), reference(H))


if __name__ == '__main__':
    print('Testing PennyLane version ' + qml.version() + ', default.qubit plugin.')
    # run the tests in this file
    suite = unittest.TestSuite()
    for t in (TestAuxillaryFunctions,
              TestStateFunctions,
              TestDefaultQubitDevice,
              TestDefaultQubitIntegration):
        ttt = unittest.TestLoader().loadTestsFromTestCase(t)
        suite.addTests(ttt)
    unittest.TextTestRunner().run(suite)



"""Variational quantum classifier

This example shows that a variational quantum classifier
can be optimized to reproduce the parity function.

"""

import pennylane as qml
from pennylane import numpy as np
from pennylane.optimize import NesterovMomentumOptimizer

dev = qml.device('default.qubit', wires=4)


def layer(W):
    """ Single layer of the variational classifier.

    Args:
        W: array of variables
    """

    qml.Rot(W[0, 0], W[0, 1], W[0, 2], wires=0)
    qml.Rot(W[1, 0], W[1, 1], W[1, 2], wires=1)
    qml.Rot(W[2, 0], W[2, 1], W[2, 2], wires=2)
    qml.Rot(W[3, 0], W[3, 1], W[3, 2], wires=3)

    qml.CNOT(wires=[0, 1])
    qml.CNOT(wires=[1, 2])
    qml.CNOT(wires=[2, 3])
    qml.CNOT(wires=[3, 0])


def statepreparation(x):
    """ Encodes data input x into quantum state.

    Args:
        x: single input vector
    """

    qml.BasisState(x, wires=[0, 1, 2, 3])


@qml.qnode(dev)
def circuit(weights, x=None):
    """The circuit of the variational classifier.

    Args:
        weights (array[float]): array of variables
        x: single input vector

    Returns:
        expectation of Pauli-Z operator on Qubit 0
    """

    statepreparation(x)

    for W in weights:
        layer(W)

    return qml.expval.PauliZ(0)


def variational_classifier(var, x=None):
    """The variational classifier.

    Args:
        var (array[float]): array of variables
        x: single input vector

    Returns:
        continuous output of the model
    """

    weights = var[0]
    bias = var[1]

    return circuit(weights, x=x) + bias


def square_loss(labels, predictions):
    """ Square loss function

    Args:
        labels (array[float]): 1-d array of labels
        predictions (array[float]): 1-d array of predictions

    Returns:
        float: square loss
    """
    loss = 0
    for l, p in zip(labels, predictions):
        loss += (l - p) ** 2
    loss = loss / len(labels)

    return loss


def accuracy(labels, predictions):
    """ Share of equal labels and predictions

    Args:
        labels (array[float]): 1-d array of labels
        predictions (array[float]): 1-d array of predictions

    Returns:
        float: accuracy
    """

    loss = 0
    for l, p in zip(labels, predictions):
        if abs(l - p) < 1e-5:
            loss += 1
    loss = loss / len(labels)

    return loss


def cost(var, X, Y):
    """Cost (error) function to be minimized."""

    predictions = [variational_classifier(var, x=x) for x in X]

    return square_loss(Y, predictions)


# load parity data
data = np.loadtxt("data/parity.txt")
X = data[:, :-1]
Y = data[:, -1]
Y = Y * 2 - np.ones(len(Y))  # shift label from {0, 1} to {-1, 1}

# initialize weight layers
np.random.seed(0)
num_qubits = 4
num_layers = 2
var_init = (0.01 * np.random.randn(num_layers, num_qubits, 3), 0.0)

# create optimizer
opt = NesterovMomentumOptimizer(0.5)
batch_size = 5

# train the variational classifier
var = var_init
for it in range(25):

    # Update the weights by one optimizer step
    batch_index = np.random.randint(0, len(X), (batch_size, ))
    X_batch = X[batch_index]
    Y_batch = Y[batch_index]
    var = opt.step(lambda v: cost(v, X_batch, Y_batch), var)

    # Compute accuracy
    predictions = [np.sign(variational_classifier(var, x=x)) for x in X]
    acc = accuracy(Y, predictions)

    print("Iter: {:5d} | Cost: {:0.7f} | Accuracy: {:0.7f} ".format(it + 1, cost(var, X, Y), acc))



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. _qubit_ops:

Qubit quantum operations
========================

.. currentmodule:: pennylane.ops.qubit

**Module name:** :mod:`pennylane.ops.qubit`

This section contains the available built-in discrete-variable
quantum operations supported by PennyLane, as well as their conventions.

Gates
-----

.. autosummary::
    Hadamard
    PauliX
    PauliY
    PauliZ
    CNOT
    CZ
    SWAP
    RX
    RY
    RZ
    PhaseShift
    Rot
    QubitUnitary


State preparation
-----------------

.. autosummary::
    BasisState
    QubitStateVector


Code details
~~~~~~~~~~~~
"""

from pennylane.operation import Operation


class Hadamard(Operation):
    r"""Hadamard(wires)
    The Hadamard operator

    .. math:: H = \frac{1}{\sqrt{2}}\begin{bmatrix} 1 & 1\\ 1 & -1\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class PauliX(Operation):
    r"""PauliX(wires)
    The Pauli X operator

    .. math:: \sigma_x = \begin{bmatrix} 0 & 1 \\ 1 & 0\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class PauliY(Operation):
    r"""PauliY(wires)
    The Pauli Y operator

    .. math:: \sigma_y = \begin{bmatrix} 0 & -i \\ i & 0\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class PauliZ(Operation):
    r"""PauliZ(wires)
    The Pauli Z operator

    .. math:: \sigma_z = \begin{bmatrix} 1 & 0 \\ 0 & -1\end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 0
    num_wires = 1
    par_domain = None


class CNOT(Operation):
    r"""CNOT(wires)
    The controlled-NOT operator

    .. math:: CNOT = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & 0 & 1\\
            0 & 0 & 1 & 0
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wires the operation acts on
    """
    num_params = 0
    num_wires = 2
    par_domain = None


class CZ(Operation):
    r"""CZ(wires)
    The controlled-Z operator

    .. math:: CZ = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0\\
            0 & 0 & 1 & 0\\
            0 & 0 & 0 & -1
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wires the operation acts on
    """
    num_params = 0
    num_wires = 2
    par_domain = None


class SWAP(Operation):
    r"""SWAP(wires)
    The swap operator

    .. math:: SWAP = \begin{bmatrix}
            1 & 0 & 0 & 0 \\
            0 & 0 & 1 & 0\\
            0 & 1 & 0 & 0\\
            0 & 0 & 0 & 1
        \end{bmatrix}.

    .. note:: The first wire provided corresponds to the **control qubit**.

    **Details:**

    * Number of wires: 2
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wires the operation acts on
    """
    num_params = 0
    num_wires = 2
    par_domain = None


class RX(Operation):
    r"""RX(phi, wires)
    The single qubit X rotation

    .. math:: R_x(\phi) = e^{-i\phi\sigma_x/2} = \begin{bmatrix}
                \cos(\phi/2) & -i\sin(\phi/2) \\
                -i\sin(\phi/2) & \cos(\phi/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}R_x(\phi) = \frac{1}{2}\left[R_x(\phi+\pi/2)+R_x(\phi-\pi/2)\right]`

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = 'R'
    grad_method = 'A'


class RY(Operation):
    r"""RY(phi, wires)
    The single qubit Y rotation

    .. math:: R_y(\phi) = e^{-i\phi\sigma_y/2} = \begin{bmatrix}
                \cos(\phi/2) & -\sin(\phi/2) \\
                \sin(\phi/2) & \cos(\phi/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}R_y(\phi) = \frac{1}{2}\left[R_y(\phi+\pi/2)+R_y(\phi-\pi/2)\right]`

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = 'R'
    grad_method = 'A'


class RZ(Operation):
    r"""RZ(phi, wires)
    The single qubit Z rotation

    .. math:: R_z(\phi) = e^{-i\phi\sigma_z/2} = \begin{bmatrix}
                e^{-i\phi/2} & 0 \\
                0 & e^{i\phi/2}
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}R_z(\phi) = \frac{1}{2}\left[R_z(\phi+\pi/2)+R_z(\phi-\pi/2)\right]`

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = 'R'
    grad_method = 'A'


class PhaseShift(Operation):
    r"""PhaseShift(phi, wires)
    Arbitrary single qubit local phase shift

    .. math:: R_\phi(\phi) = e^{i\phi/2}R_z(\phi) = \begin{bmatrix}
                1 & 0 \\
                0 & e^{i\phi}
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}R_\phi(\phi) = \frac{1}{2}\left[R_\phi(\phi+\pi/2)+R_\phi(\phi-\pi/2)\right]`

    Args:
        phi (float): rotation angle :math:`\phi`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 1
    num_wires = 1
    par_domain = 'R'
    grad_method = 'A'


class Rot(Operation):
    r"""Rot(phi, theta, omega, wires)
    Arbitrary single qubit rotation

    .. math:: R(\phi,\theta,\omega) = RZ(\omega)RY(\theta)RZ(\phi)= \begin{bmatrix}
                e^{-i(\phi+\omega)/2}\cos(\theta/2) & -e^{i(\phi-\omega)/2}\sin(\theta/2) \\
                e^{-i(\phi-\omega)/2}\sin(\theta/2) & e^{i(\phi+\omega)/2}\cos(\theta/2)
            \end{bmatrix}.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 1
    * Gradient recipe: :math:`\frac{d}{d\phi}R(\phi) = \frac{1}{2}\left[R(\phi+\pi/2)+R(\phi-\pi/2)\right]`.
      This gradient recipe applies for each angle argument :math:`\{\phi, \theta, \omega\}`.

    Args:
        phi (float): rotation angle :math:`\phi`
        theta (float): rotation angle :math:`\theta`
        omega (float): rotation angle :math:`\omega`
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_params = 3
    num_wires = 1
    par_domain = 'R'
    grad_method = 'A'


#=============================================================================
# Arbitrary operations
#=============================================================================


class QubitUnitary(Operation):
    r"""QubitUnitary(U, wires)
    Apply an arbitrary unitary matrix

    **Details:**

    * Number of wires: None (applied to the entire system)
    * Number of parameters: 1
    * Gradient recipe: None (uses finite difference)

    Args:
        U (array[complex]): square unitary matrix
        wires (Sequence[int] or int): the wire(s) the operation acts on
    """
    num_params = 1
    num_wires = 0
    par_domain = 'A'
    grad_method = 'F'


#=============================================================================
# State preparation
#=============================================================================


class BasisState(Operation):
    r"""BasisState(n, wires)
    Prepares a single computational basis state.

    **Details:**

    * Number of wires: None (applied to the entire system)
    * Number of parameters: 1
    * Gradient recipe: None (integer parameters not supported)

    Args:
        n (array): prepares the basis state :math:`\ket{n}`, where ``n`` is an
            array of integers from the set :math:`\{0, 1\}`, i.e.,
            if ``n = np.array([0, 1, 0])``, prepares the state :math:`|010\rangle`.
        wires (Sequence[int] or int): the wire(s) the operation acts on
    """
    num_params = 1
    num_wires = 0
    par_domain = 'A'
    grad_method = None


class QubitStateVector(Operation):
    r"""QubitStateVector(state, wires)
    Prepare subsystems using the given ket vector in the Fock basis.

    **Details:**

    * Number of wires: None (applied to the entire system)
    * Number of parameters: 1
    * Gradient recipe: None (uses finite difference)

    Args:
        state (array[complex]): a state vector of size 2**len(wires)
        wires (Sequence[int] or int): the wire(s) the operation acts on
    """
    num_params = 1
    num_wires = 0
    par_domain = 'A'
    grad_method = 'F'


all_ops = [
    Hadamard,
    PauliX,
    PauliY,
    PauliZ,
    CNOT,
    CZ,
    SWAP,
    RX,
    RY,
    RZ,
    PhaseShift,
    Rot,
    BasisState,
    QubitStateVector,
    QubitUnitary
]


__all__ = [cls.__name__ for cls in all_ops]



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. _qubit_expval:

Qubit quantum expectations
==========================

.. currentmodule:: pennylane.expval.qubit

**Module name:** :mod:`pennylane.expval.qubit`

.. role:: html(raw)
   :format: html

This section contains the available built-in discrete-variable
quantum operations supported by PennyLane, as well as their conventions.

.. note:: Currently, all expectation commands return scalars.

:html:`<h3>Summary</h3>`

.. autosummary::
    PauliX
    PauliY
    PauliZ
    Hadamard
    Hermitian
    Identity

:html:`<h3>Code details</h3>`
"""

from pennylane.operation import Expectation


class PauliX(Expectation):
    r"""pennylane.expval.PauliX(wires)
    Expectation value of :class:`PauliX<pennylane.ops.qubit.PauliX>`.

    This expectation command returns the value

    .. math::
        \braket{\sigma_z} = \braketT{\psi}{\cdots \otimes I\otimes \sigma_x\otimes I\cdots}{\psi}

    where :math:`\sigma_x` acts on the requested wire.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_wires = 1
    num_params = 0
    par_domain = None


class PauliY(Expectation):
    r"""pennylane.expval.PauliY(wires)
    Expectation value of :class:`PauliY<pennylane.ops.qubit.PauliY>`.

    This expectation command returns the value

    .. math::
        \braket{\sigma_z} = \braketT{\psi}{\cdots \otimes I\otimes \sigma_y\otimes I\cdots}{\psi}

    where :math:`\sigma_y` acts on the requested wire

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_wires = 1
    num_params = 0
    par_domain = None


class PauliZ(Expectation):
    r"""pennylane.expval.PauliZ(wires)
    Expectation value of :class:`PauliZ<pennylane.ops.qubit.PauliZ>`.

    This expectation command returns the value

    .. math::
        \braket{\sigma_z} = \braketT{\psi}{\cdots \otimes I\otimes \sigma_z\otimes I\cdots}{\psi}

    where :math:`\sigma_z` acts on the requested wire.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_wires = 1
    num_params = 0
    par_domain = None


class Hadamard(Expectation):
    r"""pennylane.expval.Hadamard(wires)
    Expectation value of the :class:`Hadamard<pennylane.ops.qubit.Hadamard>` observable.

    This expectation command returns the value

    .. math::
        \braket{H} = \braketT{\psi}{\cdots \otimes I\otimes H\otimes I\cdots}{\psi}

    where :math:`H` acts on the requested wire.

    **Details:**

    * Number of wires: 1
    * Number of parameters: 0

    Args:
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_wires = 1
    num_params = 0
    par_domain = None


class Hermitian(Expectation):
    r"""pennylane.expval.Hermitian(A, wires)
    Expectation value of an arbitrary Hermitian observable.

    For a Hermitian matrix :math:`A`, this expectation command returns the value

    .. math::
        \braket{A} = \braketT{\psi}{\cdots \otimes I\otimes A\otimes I\cdots}{\psi}

    where :math:`A` acts on the requested wire.

    Args:
        A (array): square hermitian matrix.
        wires (Sequence[int] or int): the wire the operation acts on
    """
    num_wires = 1
    num_params = 1
    par_domain = 'A'
    grad_method = 'F'

# As both the qubit and the CV case need an Identity Expectation,
# and these need to reside in the same name space but have to have
# different types, this Identity class is not imported into expval
# directly (it is not put in __all__ below) and instead expval
# contains a placeholder class Identity that returns appropriate
# Identity instances via __new__() suitable for the respective device.
class Identity(Expectation):
    r"""pennylane.expval.Identity(wires)
    Expectation value of the identity observable :math:`\I`.

    The expectation of this observable

    .. math::
        E[\I] = \text{Tr}(\I \rho)

    corresponds to the trace of the quantum state, which in exact
    simulators should always be equal to 1.

    .. note::

        Can be used to check normalization in approximate simulators.

    """
    num_wires = 0
    num_params = 0
    par_domain = None
    grad_method = None


all_ops = [PauliX, PauliY, PauliZ, Hadamard, Hermitian]

__all__ = [cls.__name__ for cls in all_ops]



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane` configuration classe :class:`Configuration`.
"""
# pylint: disable=protected-access
import unittest
import os
import logging as log

import toml

from defaults import pennylane, BaseTest
import pennylane as qml
from pennylane import Configuration

log.getLogger('defaults')


filename = 'default_config.toml'
expected_config = toml.load(filename)


class BasicTest(BaseTest):
    """Configuration class tests."""

    def test_loading_current_directory(self):
        """Test that the default configuration file can be loaded
        from the current directory."""
        self.logTestName()

        os.curdir = "."
        os.environ["PENNYLANE_CONF"] = ""
        config = Configuration(name=filename)

        self.assertEqual(config._config, expected_config)
        self.assertEqual(config.path, os.path.join(os.curdir, filename))

    def test_loading_environment_variable(self):
        """Test that the default configuration file can be loaded
        from an environment variable."""
        self.logTestName()

        os.curdir = "None"
        os.environ["PENNYLANE_CONF"] = os.getcwd()
        config = Configuration(name=filename)

        self.assertEqual(config._config, expected_config)
        self.assertEqual(config._env_config_dir, os.environ["PENNYLANE_CONF"])
        self.assertEqual(config.path, os.path.join(os.environ["PENNYLANE_CONF"], filename))

    def test_loading_absolute_path(self):
        """Test that the default configuration file can be loaded
        from an absolute path."""
        self.logTestName()

        os.curdir = "None"
        os.environ["PENNYLANE_CONF"] = ""
        config = Configuration(name=os.path.join(os.getcwd(), filename))

        self.assertEqual(config._config, expected_config)
        self.assertEqual(config.path, os.path.join(os.getcwd(), filename))

    def test_not_found_warning(self):
        """Test that a warning is raised if no configuration file found."""
        self.logTestName()

        with self.assertLogs(level='WARNING') as l:
            Configuration('noconfig')
            self.assertEqual(len(l.output), 1)
            self.assertEqual(len(l.records), 1)
            self.assertIn('No PennyLane configuration file found.', l.output[0])

    def test_save(self):
        """Test saving a configuration file."""
        self.logTestName()

        config = Configuration(name=filename)

        # make a change
        config['strawberryfields.global']['shots'] = 10
        config.save('test_config.toml')

        result = toml.load('test_config.toml')
        os.remove('test_config.toml')
        self.assertEqual(config._config, result)

    def test_get_item(self):
        """Test getting items."""
        self.logTestName()

        config = Configuration(name=filename)

        # get existing options
        self.assertEqual(config['main.shots'], 0)
        self.assertEqual(config['main']['shots'], 0)
        self.assertEqual(config['strawberryfields.global.hbar'], 1)
        self.assertEqual(config['strawberryfields.global']['hbar'], 1)

        # get nested dictionaries
        self.assertEqual(config['strawberryfields.fock'], {'cutoff_dim': 10})

        # get key that doesn't exist
        self.assertEqual(config['projectq.ibm.idonotexist'], {})

    def test_set_item(self):
        """Test setting items."""
        self.logTestName()

        config = Configuration(name=filename)

        # set existing options
        config['main.shots'] = 10
        self.assertEqual(config['main.shots'], 10)
        self.assertEqual(config['main']['shots'], 10)

        config['strawberryfields.global']['hbar'] = 5
        self.assertEqual(config['strawberryfields.global.hbar'], 5)
        self.assertEqual(config['strawberryfields.global']['hbar'], 5)

        # set new options
        config['projectq.ibm']['device'] = 'ibmqx4'
        self.assertEqual(config['projectq.ibm.device'], 'ibmqx4')

        # set nested dictionaries
        config['strawberryfields.tf'] = {'batched': True, 'cutoff_dim': 6}
        self.assertEqual(config['strawberryfields.tf'], {'batched': True, 'cutoff_dim': 6})

        # set nested keys that don't exist dictionaries
        config['strawberryfields.another.hello.world'] = 5
        self.assertEqual(config['strawberryfields.another'], {'hello': {'world': 5}})

    def test_bool(self):
        """Test boolean value of the Configuration object."""
        self.logTestName()

        # test false if no config is loaded
        config = Configuration('noconfig')
        self.assertFalse(config)

        # test true if config is loaded
        config = Configuration(filename)
        self.assertTrue(config)


class PennyLaneInitTests(BaseTest):
    """Tests to ensure that the code in PennyLane/__init__.py
    correctly knows how to load and use configuration data"""

    def test_device_load(self):
        """Test loading a device with a configuration."""
        self.logTestName()

        config = Configuration(name=filename)
        dev = qml.device('default.gaussian', wires=2, config=config)

        self.assertTrue(dev.hbar, 1)

if __name__ == '__main__':
    print('Testing PennyLane version ' + pennylane.version() + ', Configuration class.')
    # run the tests in this file
    suite = unittest.TestSuite()
    for t in (BasicTest, PennyLaneInitTests):
        ttt = unittest.TestLoader().loadTestsFromTestCase(t)
        suite.addTests(ttt)
    unittest.TextTestRunner().run(suite)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. _library_overview:

Library overview
================

The PennyLane codebase contains a number of complementary components.
These can be roughly separated into a user-interface, supported core
operations, and a developer API.

Software components
-------------------

**User interface**

The main user-interface to PennyLane. These are the functions and
classes that will be used by a majority of users. For a good introduction
on the user-interface of PennyLane, have a look at our tutorials.

* The device loader: :func:`pennylane.device`
* The quantum node object: :mod:`pennylane.QNode <pennylane.qnode>`
* The QNode decorator: :mod:`pennylane.qnode <pennylane.decorator>`
* Optimization methods: :mod:`pennylane.optimize`
* Configuration: :mod:`pennylane.Configuration <pennylane.configuration>`
* Utilities: :mod:`pennylane.utils <pennylane.utils>`
* NumPy with support for automatic differentiation: :mod:`pennylane.numpy <pennylane.numpy>`

**Core operations**

The main operations and expectations supported by PennyLane.
Each of these operations/expectations supports a method
of automatic differentiation (either analytically or numerically).

The conventions used in defining these operations are also
provided here.

* Supported operations: :mod:`pennylane.ops`
* Supported expectations: :mod:`pennylane.expval`

**Developer API**

Used to develop new plugins for PennyLane - providing new devices
for QNodes, or supporting new operations and expectations. For more
details, see :ref:`developer_overview`.

* The base Device class: :mod:`pennylane.Device <pennylane._device>`
* Symbolic quantum operations: :mod:`pennylane.operation`
* Quantum circuit parameters: :mod:`pennylane.variable`

Summary
-------

.. autosummary::


    ~configuration.Configuration
    ~_device.Device
    ~_device.DeviceError
    device
    expval
    grad
    jacobian
    ~autograd.numpy
    ops
    optimize
    ~qnode.QNode
    ~decorator.qnode
    ~qnode.QuantumFunctionError
    version

.. note::

    All individual operations (contained in :mod:`~.ops`) and optimizers
    (contained in :mod:`~.optimize`) may also be imported directly from PennyLane.
    Expectation values, however, must be accessed via the :mod:`~.expval` module.

Code details
~~~~~~~~~~~~
"""
import os
import logging as log
from pkg_resources import iter_entry_points

from autograd import numpy
from autograd import grad as _grad
from autograd import jacobian as _jacobian

from semantic_version import Version, Spec

import pennylane.operation
import pennylane.expval
import pennylane.template


from .configuration import Configuration
from ._device import Device, DeviceError
from .ops import *
from .optimize import *
from .qnode import QNode, QuantumFunctionError
from ._version import __version__

# NOTE: this has to be imported last,
# otherwise it will clash with the .qnode import.
from .decorator import qnode


# overwrite module docstrings
numpy.__doc__ = "NumPy with automatic differentiation support, provided by Autograd."


# set up logging
numeric_level = 100 # info
if "LOGGING" in os.environ:
    logLevel = os.environ["LOGGING"]
    numeric_level = getattr(log, logLevel.upper(), 10)


log.basicConfig(
    level=numeric_level,
    format='%(asctime)s %(levelname)s %(message)s',
    datefmt='%H:%M:%S'
)
log.captureWarnings(True)


# Look for an existing configuration file
default_config = Configuration('config.toml')


# get list of installed plugin devices
plugin_devices = {
    entry.name: entry for entry in iter_entry_points('pennylane.plugins')
}


def device(name, *args, **kwargs):
    r"""device(name, wires=1, *args, **kwargs)
    Load a plugin :class:`~.Device` and return the instance.

    This function is used to load a particular quantum device,
    which can then be used to construct QNodes.

    PennyLane comes with support for the following two devices:

    * :mod:`'default.qubit' <pennylane.plugins.default_qubit>`: a simple pure
      state simulator of qubit-based quantum circuit architectures.

    * :mod:`'default.gaussian' <pennylane.plugins.default_gaussian>`: a simple simulator
      of Gaussian states and operations on continuous-variable circuit architectures.

    In addition, additional devices are supported through plugins — see
    :ref:`plugins` for more details.

    All devices must be loaded by specifying their **short-name** as listed above,
    followed by the number of *wires* (subsystems) you wish to initialize.

    Some devices may accept additional arguments. For instance,
    ``default.gaussian`` accepts the keyword argument ``hbar``, to set
    the convention used in the commutation relation :math:`[\x,\p]=i\hbar`
    (by default set to 2).

    Please refer to the documentation for the individual devices to see any
    additional arguments that might be required or supported.

    Args:
        name (str): the name of the device to load
        wires (int): the number of wires (subsystems) to initialise
            the device with

    Keyword Args:
        config (pennylane.Configuration): an PennyLane configuration object
            that contains global and/or device specific configurations.
    """
    if name in plugin_devices:
        options = {}

        # load global configuration settings if available
        config = kwargs.get('config', default_config)

        if config:
            # combine configuration options with keyword arguments.
            # Keyword arguments take preference, followed by device options,
            # followed by plugin options, followed by global options.
            options.update(config['main'])
            options.update(config[name.split('.')[0]+'.global'])
            options.update(config[name])

        kwargs.pop("config", None)
        options.update(kwargs)

        # loads the plugin device class
        plugin_device_class = plugin_devices[name].load()

        if Version(version()) not in Spec(plugin_device_class.pennylane_requires):
            raise DeviceError("The {} plugin requires PennyLane versions {}, however PennyLane "
                              "version {} is installed.".format(name, plugin_device_class.pennylane_requires, __version__))

        # load plugin device
        return plugin_device_class(*args, **options)

    raise DeviceError('Device does not exist. Make sure the required plugin is installed.')


def grad(func, argnum):
    """Returns the gradient (as a callable function) of functions accessible within PennyLane.

    This is a wrapper around the :mod:`autograd.grad` function.

    Args:
        func (function): a Python function or QNode that contains
            a combination of quantum and classical nodes
        argnum (int or list(int)): which argument(s) to take the gradient
            with respect to

    Returns:
        function: the function that returns the gradient of the input
        function with respect to the arguments in argnum
    """
    # pylint: disable=no-value-for-parameter
    return _grad(func, argnum)


def jacobian(func, argnum):
    """Returns the Jacobian (as a callable function) of vector-valued
    functions accessible within PennyLane.

    This is a wrapper around the :mod:`autograd.jacobian` function.

    Args:
        func (function): a vector-valued Python function or QNode that contains
            a combination of quantum and classical nodes. The output of the computation
            must consist of a single NumPy array (if classical) or a tuple of
            expectation values (if a quantum node)
        argnum (int): which argument to take the gradient
            with respect to. If the argument is a NumPy array, then the Jacobian
            corresponding to all input elements and all output elements is returned.

    Returns:
        function: the function that returns the Jacobian of the input
        function with respect to the arguments in argnum
    """
    # pylint: disable=no-value-for-parameter
    return _jacobian(func, argnum)


def version():
    """Returns the PennyLane version number."""
    return __version__



"""Continuous-variable quantum neural network example.

In this demo we implement the photonic quantum neural net model
from Killoran et al. (arXiv:1806.06871) with the example
of function fitting.
"""

import pennylane as qml
from pennylane import numpy as np
from pennylane.optimize import AdamOptimizer

try:
    dev = qml.device('strawberryfields.fock', wires=1, cutoff_dim=10)    
except:
    print("To run this demo you need to install the strawberryfields plugin...")


def layer(v):
    """ Single layer of the quantum neural net.

    Args:
        v (array[float]): array of variables for one layer
    """
    # Matrix multiplication of input layer
    qml.Rotation(v[0], wires=0)
    qml.Squeezing(v[1], 0., wires=0)
    qml.Rotation(v[2], wires=0)

    # Bias
    qml.Displacement(v[3], 0., wires=0)

    # Element-wise nonlinear transformation
    qml.Kerr(v[4], wires=0)


@qml.qnode(dev)
def quantum_neural_net(var, x=None):
    """The quantum neural net variational circuit.

    Args:
        var (array[float]): array of variables
        x (array[float]): single input vector

    Returns:
        float: expectation of Homodyne measurement on Mode 0
    """
    # Encode input x into quantum state
    qml.Displacement(x, 0., wires=0)

    # "layer" subcircuits
    for v in var:
        layer(v)

    return qml.expval.X(0)


def square_loss(labels, predictions):
    """ Square loss function

    Args:
        labels (array[float]): 1-d array of labels
        predictions (array[float]): 1-d array of predictions

    Returns:
        float: square loss
    """
    loss = 0
    for l, p in zip(labels, predictions):
        loss = loss + (l - p) ** 2
    loss = loss / len(labels)

    return loss


def cost(var, features, labels):
    """Cost function to be minimized.

    Args:
        var (array[float]): array of variables
        features (array[float]): 2-d array of input vectors
        labels (array[float]): 1-d array of targets

    Returns:
        float: loss
    """
    # Compute prediction for each input in data batch
    preds = [quantum_neural_net(var, x=x) for x in features]

    return square_loss(labels, preds)


# load function data
data = np.loadtxt("data/sine.txt")
X = data[:, 0]
Y = data[:, 1]

# initialize weights
np.random.seed(0)
num_layers = 4
var_init = 0.05 * np.random.randn(num_layers, 5)

# create optimizer
opt = AdamOptimizer(0.01, beta1=0.9, beta2=0.999)

# train
var = var_init
for it in range(500):
    var = opt.step(lambda v: cost(v, X, Y), var)
    
    print("Iter: {:5d} | Cost: {:0.7f} ".format(it + 1, cost(var, X, Y)))



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.interface.tfe` QNode interface.
"""

import unittest
import logging as log
log.getLogger('defaults')

import numpy as np

try:
    import tensorflow as tf
    import tensorflow.contrib.eager as tfe
    tf.enable_eager_execution()
    tf_support = True
except ImportError as e:
    tf_support = False


from defaults import pennylane as qml, BaseTest

from pennylane.qnode import _flatten, unflatten, QNode, QuantumFunctionError
from pennylane.plugins.default_qubit import CNOT, Rotx, Roty, Rotz, I, Y, Z
from pennylane._device import DeviceError


def expZ(state):
    return np.abs(state[0]) ** 2 - np.abs(state[1]) ** 2


class TFEQNodeTests(BaseTest):
    """TFEQNode basic tests."""
    def setUp(self):
        if not tf_support:
            self.skipTest('TFE interface not tested')

        self.dev1 = qml.device('default.qubit', wires=1)
        self.dev2 = qml.device('default.qubit', wires=2)

    def test_qnode_fail(self):
        """Tests that QNode initialization failures correctly raise exceptions."""
        self.logTestName()
        par = tfe.Variable(0.5)

        #---------------------------------------------------------
        ## faulty quantum functions

        # qfunc must return only Expectations
        @qml.qnode(self.dev2, interface='tfe')
        def qf(x):
            qml.RX(x, [0])
            return qml.expval.PauliZ(0), 0.3

        with self.assertRaisesRegex(QuantumFunctionError, 'must return either'):
            qf(par)

        # all EVs must be returned...
        @qml.qnode(self.dev2, interface='tfe')
        def qf(x):
            qml.RX(x, [0])
            ex = qml.expval.PauliZ(1)
            return qml.expval.PauliZ(0)

        with self.assertRaisesRegex(QuantumFunctionError, 'All measured expectation values'):
            qf(par)

        # ...in the correct order
        @qml.qnode(self.dev2, interface='tfe')
        def qf(x):
            qml.RX(x, [0])
            ex = qml.expval.PauliZ(1)
            return qml.expval.PauliZ(0), ex

        with self.assertRaisesRegex(QuantumFunctionError, 'All measured expectation values'):
            qf(par)

        # gates must precede EVs
        @qml.qnode(self.dev2, interface='tfe')
        def qf(x):
            qml.RX(x, [0])
            ev = qml.expval.PauliZ(1)
            qml.RY(0.5, [0])
            return ev

        with self.assertRaisesRegex(QuantumFunctionError, 'gates must precede'):
            qf(par)

        # a wire cannot be measured more than once
        @qml.qnode(self.dev2, interface='tfe')
        def qf(x):
            qml.RX(x, [0])
            qml.CNOT([0, 1])
            return qml.expval.PauliZ(0), qml.expval.PauliZ(1), qml.expval.PauliX(0)

        with self.assertRaisesRegex(QuantumFunctionError, 'can only be measured once'):
            qf(par)

        # device must have enough wires for the qfunc
        @qml.qnode(self.dev2, interface='tfe')
        def qf(x):
            qml.RX(x, [0])
            qml.CNOT([0, 2])
            return qml.expval.PauliZ(0)

        with self.assertRaisesRegex(QuantumFunctionError, 'applied to invalid wire'):
            qf(par)

        # CV and discrete ops must not be mixed
        @qml.qnode(self.dev1, interface='tfe')
        def qf(x):
            qml.RX(x, [0])
            qml.Displacement(0.5, 0, [0])
            return qml.expval.PauliZ(0)

        with self.assertRaisesRegex(QuantumFunctionError, 'Continuous and discrete'):
            qf(par)

        # default plugin cannot execute CV operations, neither gates...
        @qml.qnode(self.dev1, interface='tfe')
        def qf(x):
            qml.Displacement(0.5, 0, [0])
            return qml.expval.X(0)

        with self.assertRaisesRegex(DeviceError, 'Gate [a-zA-Z]+ not supported on device'):
            qf(par)

        # ...nor observables
        @qml.qnode(self.dev1, interface='tfe')
        def qf(x):
            return qml.expval.X(0)

        with self.assertRaisesRegex(DeviceError, 'Expectation [a-zA-Z]+ not supported on device'):
            qf(par)

    def test_qnode_fanout(self):
        """Tests that qnodes can compute the correct function when the same parameter is used in multiple gates."""
        self.logTestName()

        @qml.qnode(self.dev1, interface='tfe')
        def circuit(reused_param, other_param):
            qml.RX(reused_param, [0])
            qml.RZ(other_param, [0])
            qml.RX(reused_param, [0])
            return qml.expval.PauliZ(0)

        thetas = tf.linspace(-2*np.pi, 2*np.pi, 7)

        for reused_param in thetas:
            for theta in thetas:
                other_param = theta ** 2 / 11
                y_eval = circuit(reused_param, other_param)
                Rx = Rotx(reused_param.numpy())
                Rz = Rotz(other_param.numpy())
                zero_state = np.array([1.,0.])
                final_state = (Rx @ Rz @ Rx @ zero_state)
                y_true = expZ(final_state)
                self.assertAlmostEqual(y_eval, y_true, delta=self.tol)

    def test_qnode_array_parameters(self):
        "Test that QNode can take arrays as input arguments, and that they interact properly with TensorFlow."
        self.logTestName()

        @qml.qnode(self.dev1, interface='tfe')
        def circuit_n1s(dummy1, array, dummy2):
            qml.RY(0.5 * array[0,1], 0)
            qml.RY(-0.5 * array[1,1], 0)
            return qml.expval.PauliX(0)  # returns a scalar

        @qml.qnode(self.dev1, interface='tfe')
        def circuit_n1v(dummy1, array, dummy2):
            qml.RY(0.5 * array[0,1], 0)
            qml.RY(-0.5 * array[1,1], 0)
            return qml.expval.PauliX(0),  # note the comma, returns a 1-vector

        @qml.qnode(self.dev2, interface='tfe')
        def circuit_nn(dummy1, array, dummy2):
            qml.RY(0.5 * array[0,1], 0)
            qml.RY(-0.5 * array[1,1], 0)
            qml.RY(array[1,0], 1)
            return qml.expval.PauliX(0), qml.expval.PauliX(1)  # returns a 2-vector

        grad_target = (np.array(1.), np.array([[0.5,  0.43879, 0], [0, -0.43879, 0]]), np.array(-0.4))
        cost_target = 1.03257

        for circuit in [circuit_n1s, circuit_n1v, circuit_nn]:

            args = (tfe.Variable(0.46), tfe.Variable([[2., 3., 0.3], [7., 4., 2.1]]), tfe.Variable(-0.13))

            def cost(x, array, y):
                c = tf.cast(circuit(tf.constant(0.111), array, tf.constant(4.5)), tf.float32)
                if c.shape != tf.TensorShape([]):
                    c = c[0]  # get a scalar
                return c +0.5*array[0,0] +x -0.4*y

            with tf.GradientTape() as tape:
                cost_res = cost(*args)
                grad_res = np.array([i.numpy() for i in tape.gradient(cost_res, [args[0], args[2]])])

            self.assertAllAlmostEqual(cost_res.numpy(), cost_target, delta=self.tol)
            self.assertAllAlmostEqual(grad_res, np.fromiter(grad_target[::2], dtype=np.float32), delta=self.tol)

    def test_array_parameters_evaluate(self):
        "Test that array parameters gives same result as positional arguments."
        self.logTestName()

        a, b, c = tf.constant(0.5), tf.constant(0.54), tf.constant(0.3)

        def ansatz(x, y, z):
            qml.QubitStateVector(np.array([1, 0, 1, 1])/np.sqrt(3), [0, 1])
            qml.Rot(x, y, z, 0)
            qml.CNOT([0, 1])
            return qml.expval.PauliZ(0), qml.expval.PauliY(1)

        @qml.qnode(self.dev2, interface='tfe')
        def circuit1(x, y, z):
            return ansatz(x, y, z)

        @qml.qnode(self.dev2, interface='tfe')
        def circuit2(x, array):
            return ansatz(x, array[0], array[1])

        @qml.qnode(self.dev2, interface='tfe')
        def circuit3(array):
            return ansatz(*array)

        positional_res = circuit1(a, b, c)
        array_res1 = circuit2(a, tfe.Variable([b, c]))
        array_res2 = circuit3(tfe.Variable([a, b, c]))
        self.assertAllAlmostEqual(positional_res, array_res1, delta=self.tol)
        self.assertAllAlmostEqual(positional_res, array_res2, delta=self.tol)

    def test_multiple_expectation_different_wires(self):
        "Tests that qnodes return multiple expectation values."
        self.logTestName()

        a, b, c = tf.constant(0.5), tf.constant(0.54), tf.constant(0.3)

        @qml.qnode(self.dev2, interface='tfe')
        def circuit(x, y, z):
            qml.RX(x, [0])
            qml.RZ(y, [0])
            qml.CNOT([0, 1])
            qml.RY(y, [0])
            qml.RX(z, [0])
            return qml.expval.PauliY(0), qml.expval.PauliZ(1)

        res = circuit(a, b, c)

        out_state = np.kron(Rotx(c.numpy()), I) @ np.kron(Roty(b.numpy()), I) @ CNOT \
            @ np.kron(Rotz(b.numpy()), I) @ np.kron(Rotx(a.numpy()), I) @ np.array([1, 0, 0, 0])

        ex0 = np.vdot(out_state, np.kron(Y, I) @ out_state)
        ex1 = np.vdot(out_state, np.kron(I, Z) @ out_state)
        ex = np.array([ex0, ex1])
        self.assertAllAlmostEqual(ex, res.numpy(), delta=self.tol)

    def test_multiple_keywordargs_used(self):
        "Tests that qnodes use multiple keyword arguments."
        self.logTestName()

        def circuit(w, x=None, y=None):
            qml.RX(x, [0])
            qml.RX(y, [1])
            return qml.expval.PauliZ(0), qml.expval.PauliZ(1)

        circuit = qml.QNode(circuit, self.dev2).to_tfe()

        c = circuit(tf.constant(1.), x=np.pi, y=np.pi)
        self.assertAllAlmostEqual(c.numpy(), [-1., -1.], delta=self.tol)

    def test_multidimensional_keywordargs_used(self):
        "Tests that qnodes use multi-dimensional keyword arguments."
        self.logTestName()

        def circuit(w, x=None):
            qml.RX(x[0], [0])
            qml.RX(x[1], [1])
            return qml.expval.PauliZ(0), qml.expval.PauliZ(1)

        circuit = qml.QNode(circuit, self.dev2).to_tfe()

        c = circuit(tf.constant(1.), x=[np.pi, np.pi])
        self.assertAllAlmostEqual(c, [-1., -1.], delta=self.tol)

    def test_keywordargs_for_wires(self):
        "Tests that wires can be passed as keyword arguments."
        self.logTestName()

        default_q = 0

        def circuit(x, q=default_q):
            qml.RY(x, wires=0)
            return qml.expval.PauliZ(q)

        circuit = qml.QNode(circuit, self.dev2).to_tfe()

        c = circuit(tf.constant(np.pi), q=1)
        self.assertAlmostEqual(c, 1., delta=self.tol)

        c = circuit(tf.constant(np.pi))
        self.assertAlmostEqual(c, -1., delta=self.tol)

    def test_keywordargs_used(self):
        "Tests that qnodes use keyword arguments."
        self.logTestName()

        def circuit(w, x=None):
            qml.RX(x, [0])
            return qml.expval.PauliZ(0)

        circuit = qml.QNode(circuit, self.dev1).to_tfe()

        c = circuit(tf.constant(1.), x=np.pi)
        self.assertAlmostEqual(c, -1., delta=self.tol)

    def test_mixture_numpy_tensors(self):
        "Tests that qnodes work with python types and tensors."
        self.logTestName()

        @qml.qnode(self.dev2, interface='tfe')
        def circuit(w, x, y):
            qml.RX(x, [0])
            qml.RX(y, [1])
            return qml.expval.PauliZ(0), qml.expval.PauliZ(1)

        c = circuit(tf.constant(1.), np.pi, np.pi).numpy()
        self.assertAllAlmostEqual(c, [-1., -1.], delta=self.tol)

    def test_keywordarg_updated_in_multiple_calls(self):
        "Tests that qnodes update keyword arguments in consecutive calls."
        self.logTestName()

        def circuit(w, x=None):
            qml.RX(w, [0])
            qml.RX(x, [1])
            return qml.expval.PauliZ(0), qml.expval.PauliZ(1)

        circuit = qml.QNode(circuit, self.dev2).to_tfe()

        c1 = circuit(tf.constant(0.1), x=tf.constant(0.))
        c2 = circuit(tf.constant(0.1), x=np.pi)
        self.assertTrue(c1[1] != c2[1])

    def test_keywordarg_passes_through_classicalnode(self):
        "Tests that qnodes' keyword arguments pass through classical nodes."
        self.logTestName()

        def circuit(w, x=None):
            qml.RX(w, [0])
            qml.RX(x, [1])
            return qml.expval.PauliZ(0), qml.expval.PauliZ(1)

        circuit = qml.QNode(circuit, self.dev2).to_tfe()

        def classnode(w, x=None):
            return circuit(w, x=x)

        c = classnode(tf.constant(0.), x=np.pi)
        self.assertAllAlmostEqual(c, [1., -1.], delta=self.tol)

    def test_keywordarg_gradient(self):
        "Tests that qnodes' keyword arguments work with gradients"
        self.logTestName()

        def circuit(x, y, input_state=np.array([0, 0])):
            qml.BasisState(input_state, wires=[0, 1])
            qml.RX(x, wires=[0])
            qml.RY(y, wires=[0])
            return qml.expval.PauliZ(0)

        circuit = qml.QNode(circuit, self.dev2).to_tfe()

        x = 0.543
        y = 0.45632
        expected_grad = np.array([np.sin(x)*np.cos(y), np.sin(y)*np.cos(x)])

        x_t = tfe.Variable(x)
        y_t = tfe.Variable(y)

        # test first basis state against analytic result
        with tf.GradientTape() as tape:
            c = circuit(x_t, y_t, input_state=np.array([0, 0]))
            grads = np.array(tape.gradient(c, [x_t, y_t]))

        self.assertAllAlmostEqual(grads, -expected_grad, delta=self.tol)

        # test third basis state against analytic result
        with tf.GradientTape() as tape:
            c = circuit(x_t, y_t, input_state=np.array([1, 0]))
            grads = np.array(tape.gradient(c, [x_t, y_t]))

        self.assertAllAlmostEqual(grads, expected_grad, delta=self.tol)

        # test first basis state via the default keyword argument against analytic result
        with tf.GradientTape() as tape:
            c = circuit(x_t, y_t)
            grads = np.array(tape.gradient(c, [x_t, y_t]))

        self.assertAllAlmostEqual(grads, -expected_grad, delta=self.tol)


class IntegrationTests(BaseTest):
    """Integration tests to ensure the TensorFlow QNode agrees with the NumPy QNode"""

    def setUp(self):
        if not tf_support:
            self.skipTest('TFE interface not tested')

    def test_qnode_evaluation_agrees(self):
        "Tests that simple example is consistent."
        self.logTestName()

        dev = qml.device('default.qubit', wires=2)

        @qml.qnode(dev, interface='autograd')
        def circuit(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval.PauliZ(0)

        @qml.qnode(dev, interface='tfe')
        def circuit_tfe(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval.PauliZ(0)

        phi = [0.5, 0.1]
        theta = [0.2]

        phi_t = tfe.Variable(phi)
        theta_t = tfe.Variable(theta)

        autograd_eval = circuit(phi, theta)
        tfe_eval = circuit_tfe(phi_t, theta_t)
        self.assertAllAlmostEqual(autograd_eval, tfe_eval.numpy(), delta=self.tol)

    def test_qnode_gradient_agrees(self):
        "Tests that simple gradient example is consistent."
        self.logTestName()

        dev = qml.device('default.qubit', wires=2)

        @qml.qnode(dev, interface='autograd')
        def circuit(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval.PauliZ(0)

        @qml.qnode(dev, interface='tfe')
        def circuit_tfe(phi, theta):
            qml.RX(phi[0], wires=0)
            qml.RY(phi[1], wires=1)
            qml.CNOT(wires=[0, 1])
            qml.PhaseShift(theta[0], wires=0)
            return qml.expval.PauliZ(0)

        phi = [0.5, 0.1]
        theta = [0.2]

        phi_t = tfe.Variable(phi)
        theta_t = tfe.Variable(theta)

        dcircuit = qml.grad(circuit, [0, 1])
        autograd_grad = dcircuit(phi, theta)

        dcircuit = tfe.gradients_function(circuit_tfe)
        tfe_grad = dcircuit(phi_t, theta_t)

        self.assertAllAlmostEqual(autograd_grad[0], tfe_grad[0], delta=self.tol)
        self.assertAllAlmostEqual(autograd_grad[1], tfe_grad[1], delta=self.tol)


if __name__ == '__main__':
    print('Testing PennyLane version ' + qml.version() + ', QNode TFE interface.')
    # run the tests in this file
    suite = unittest.TestSuite()
    for t in (TFEQNodeTests, IntegrationTests):
        ttt = unittest.TestLoader().loadTestsFromTestCase(t)
        suite.addTests(ttt)

    unittest.TextTestRunner().run(suite)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.plugin.DefaultGaussian` device.
"""
# pylint: disable=protected-access,cell-var-from-loop
import unittest
import inspect
import logging as log

from scipy.special import factorial as fac
from scipy.linalg import block_diag

from defaults import pennylane as qml, BaseTest

from pennylane import numpy as np

from pennylane.plugins.default_gaussian import fock_prob

from pennylane.plugins.default_gaussian import (rotation, squeezing, quadratic_phase,
                                                beamsplitter, two_mode_squeezing,
                                                controlled_addition, controlled_phase)
from pennylane.plugins.default_gaussian import (vacuum_state, coherent_state,
                                                squeezed_state, displaced_squeezed_state,
                                                thermal_state)

from pennylane.plugins.default_gaussian import DefaultGaussian


log.getLogger('defaults')


U = np.array([[0.83645892-0.40533293j, -0.20215326+0.30850569j],
              [-0.23889780-0.28101519j, -0.88031770-0.29832709j]])


U2 = np.array([[-0.07843244-3.57825948e-01j, 0.71447295-5.38069384e-02j, 0.20949966+6.59100734e-05j, -0.50297381+2.35731613e-01j],
               [-0.26626692+4.53837083e-01j, 0.27771991-2.40717436e-01j, 0.41228017-1.30198687e-01j, 0.01384490-6.33200028e-01j],
               [-0.69254712-2.56963068e-02j, -0.15484858+6.57298384e-02j, -0.53082141+7.18073414e-02j, -0.41060450-1.89462315e-01j],
               [-0.09686189-3.15085273e-01j, -0.53241387-1.99491763e-01j, 0.56928622+3.97704398e-01j, -0.28671074-6.01574497e-02j]])


H = np.array([[1.02789352, 1.61296440-0.3498192j],
              [1.61296440+0.3498192j, 1.23920938+0j]])


hbar = 2

def prep_par(par, op):
    "Convert par into a list of parameters that op expects."
    if op.par_domain == 'A':
        return [np.diag([x, 1]) for x in par]
    return par


class TestAuxillaryFunctions(BaseTest):
    """Tests the auxillary functions"""

    def setUp(self):
        self.hbar = 2.

        # an arbitrary two-mode Gaussian state generated using Strawberry Fields
        self.mu = np.array([0.6862, 0.4002, 0.09, 0.558])*np.sqrt(self.hbar)
        self.cov = np.array([[0.50750512, -0.04125979, -0.21058229, -0.07866912],
                             [-0.04125979, 0.50750512, -0.07866912, -0.21058229],
                             [-0.21058229, -0.07866912, 0.95906208, 0.27133391],
                             [-0.07866912, -0.21058229, 0.27133391, 0.95906208]])*self.hbar

        # expected Fock state probabilities
        self.events = [(0, 0), (0, 1), (1, 1), (2, 3)]
        self.probs = [0.430461524043, 0.163699407559, 0.0582788388927, 0.00167706931355]

    def test_fock_prob(self):
        """Test fock_prob returns the correct Fock probabilities"""
        for idx, e in enumerate(self.events):
            res = fock_prob(self.mu, self.cov, e, hbar=self.hbar)
            self.assertAlmostEqual(res, self.probs[idx], delta=self.tol)


class TestGates(BaseTest):
    """Gate tests."""

    def test_rotation(self):
        """Test the Fourier transform of a displaced state."""
        # pylint: disable=invalid-unary-operand-type
        self.logTestName()

        alpha = 0.23+0.12j
        S = rotation(np.pi/2)

        # apply to a coherent state. F{x, p} -> {-p, x}
        out = S @ np.array([alpha.real, alpha.imag])*np.sqrt(2*hbar)
        expected = np.array([-alpha.imag, alpha.real])*np.sqrt(2*hbar)
        self.assertAllAlmostEqual(out, expected, delta=self.tol)

    def test_squeezing(self):
        """Test the squeezing symplectic transform."""
        self.logTestName()

        r = 0.543
        phi = 0.123
        S = squeezing(r, phi)

        # apply to an identity covariance matrix
        out = S @ S.T
        expected = rotation(phi/2) @ np.diag(np.exp([-2*r, 2*r])) @ rotation(phi/2).T
        self.assertAllAlmostEqual(out, expected, delta=self.tol)

    def test_quadratic_phase(self):
        """Test the quadratic phase symplectic transform."""
        self.logTestName()

        s = 0.543
        S = quadratic_phase(s)

        # apply to a coherent state. P[x, p] -> [x, p+sx]
        alpha = 0.23+0.12j
        out = S @ np.array([alpha.real, alpha.imag])*np.sqrt(2*hbar)
        expected = np.array([alpha.real, alpha.imag+s*alpha.real])*np.sqrt(2*hbar)
        self.assertAllAlmostEqual(out, expected, delta=self.tol)

    def test_beamsplitter(self):
        """Test the beamsplitter symplectic transform."""
        self.logTestName()

        theta = 0.543
        phi = 0.312
        S = beamsplitter(theta, phi)

        # apply to a coherent state. BS|a1, a2> -> |ta1-r^*a2, ra1+ta2>
        a1 = 0.23+0.12j
        a2 = 0.23+0.12j
        out = S @ np.array([a1.real, a2.real, a1.imag, a2.imag])*np.sqrt(2*hbar)

        T = np.cos(theta)
        R = np.exp(1j*phi)*np.sin(theta)
        a1out = T*a1 - R.conj()*a2
        a2out = R*a2 + T*a1
        expected = np.array([a1out.real, a2out.real, a1out.imag, a2out.imag])*np.sqrt(2*hbar)
        self.assertAllAlmostEqual(out, expected, delta=self.tol)

    def test_two_mode_squeezing(self):
        """Test the two mode squeezing symplectic transform."""
        self.logTestName()

        r = 0.543
        phi = 0.123
        S = two_mode_squeezing(r, phi)

        # test that S = B^\dagger(pi/4, 0) [S(z) x S(-z)] B(pi/4)
        B = beamsplitter(np.pi/4, 0)
        Sz = block_diag(squeezing(r, phi), squeezing(-r, phi))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]
        expected = B.conj().T @ Sz @ B
        self.assertAllAlmostEqual(S, expected, delta=self.tol)

        # test that S |a1, a2> = |ta1+ra2, ta2+ra1>
        a1 = 0.23+0.12j
        a2 = 0.23+0.12j
        out = S @ np.array([a1.real, a2.real, a1.imag, a2.imag])*np.sqrt(2*hbar)

        T = np.cosh(r)
        R = np.exp(1j*phi)*np.sinh(r)
        a1out = T*a1 + R*np.conj(a2)
        a2out = T*a2 + R*np.conj(a1)
        expected = np.array([a1out.real, a2out.real, a1out.imag, a2out.imag])*np.sqrt(2*hbar)
        self.assertAllAlmostEqual(out, expected, delta=self.tol)

    def test_controlled_addition(self):
        """Test the CX symplectic transform."""
        self.logTestName()

        s = 0.543
        S = controlled_addition(s)

        # test that S = B(theta+pi/2, 0) [S(z) x S(-z)] B(theta, 0)
        r = np.arcsinh(-s/2)
        theta = 0.5*np.arctan2(-1/np.cosh(r), -np.tanh(r))
        Sz = block_diag(squeezing(r, 0), squeezing(-r, 0))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]

        expected = beamsplitter(theta+np.pi/2, 0) @ Sz @ beamsplitter(theta, 0)
        self.assertAllAlmostEqual(S, expected, delta=self.tol)

        # test that S[x1, x2, p1, p2] -> [x1, x2+sx1, p1-sp2, p2]
        x1 = 0.5432
        x2 = -0.453
        p1 = 0.154
        p2 = -0.123
        out = S @ np.array([x1, x2, p1, p2])*np.sqrt(2*hbar)
        expected = np.array([x1, x2+s*x1, p1-s*p2, p2])*np.sqrt(2*hbar)
        self.assertAllAlmostEqual(out, expected, delta=self.tol)

    def test_controlled_phase(self):
        """Test the CZ symplectic transform."""
        self.logTestName()

        s = 0.543
        S = controlled_phase(s)

        # test that S = R_2(pi/2) CX(s) R_2(pi/2)^\dagger
        R2 = block_diag(np.identity(2), rotation(np.pi/2))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]
        expected = R2 @ controlled_addition(s) @ R2.conj().T
        self.assertAllAlmostEqual(S, expected, delta=self.tol)

        # test that S[x1, x2, p1, p2] -> [x1, x2, p1+sx2, p2+sx1]
        x1 = 0.5432
        x2 = -0.453
        p1 = 0.154
        p2 = -0.123
        out = S @ np.array([x1, x2, p1, p2])*np.sqrt(2*hbar)
        expected = np.array([x1, x2, p1+s*x2, p2+s*x1])*np.sqrt(2*hbar)
        self.assertAllAlmostEqual(out, expected, delta=self.tol)


class TestStates(BaseTest):
    """State tests."""

    def test_vacuum_state(self):
        """Test the vacuum state is correct."""
        self.logTestName()
        wires = 3
        means, cov = vacuum_state(wires, hbar=hbar)
        self.assertAllAlmostEqual(means, np.zeros([2*wires]), delta=self.tol)
        self.assertAllAlmostEqual(cov, np.identity(2*wires)*hbar/2, delta=self.tol)

    def test_coherent_state(self):
        """Test the coherent state is correct."""
        self.logTestName()
        a = 0.432-0.123j
        means, cov = coherent_state(a, hbar=hbar)
        self.assertAllAlmostEqual(means, np.array([a.real, a.imag])*np.sqrt(2*hbar), delta=self.tol)
        self.assertAllAlmostEqual(cov, np.identity(2)*hbar/2, delta=self.tol)

    def test_squeezed_state(self):
        """Test the squeezed state is correct."""
        self.logTestName()
        r = 0.432
        phi = 0.123
        means, cov = squeezed_state(r, phi, hbar=hbar)

        # test vector of means is zero
        self.assertAllAlmostEqual(means, np.zeros([2]), delta=self.tol)

        R = rotation(phi/2)
        expected = R @ np.array([[np.exp(-2*r), 0],
                                 [0, np.exp(2*r)]]) * hbar/2 @ R.T
        # test covariance matrix is correct
        self.assertAllAlmostEqual(cov, expected, delta=self.tol)

    def test_displaced_squeezed_state(self):
        """Test the displaced squeezed state is correct."""
        self.logTestName()
        alpha = 0.541+0.109j
        a = abs(alpha)
        phi_a = np.angle(alpha)
        r = 0.432
        phi_r = 0.123
        means, cov = displaced_squeezed_state(a, phi_a, r, phi_r, hbar=hbar)

        # test vector of means is correct
        self.assertAllAlmostEqual(means, np.array([alpha.real, alpha.imag])*np.sqrt(2*hbar), delta=self.tol)

        R = rotation(phi_r/2)
        expected = R @ np.array([[np.exp(-2*r), 0],
                                 [0, np.exp(2*r)]]) * hbar/2 @ R.T
        # test covariance matrix is correct
        self.assertAllAlmostEqual(cov, expected, delta=self.tol)

    def thermal_state(self):
        """Test the thermal state is correct."""
        self.logTestName()
        nbar = 0.5342
        means, cov = thermal_state(nbar, hbar=hbar)
        self.assertAllAlmostEqual(means, np.zeros([2]), delta=self.tol)
        self.assertTrue(np.all((cov.diag*2/hbar-1)/2 == nbar))



class TestDefaultGaussianDevice(BaseTest):
    """Test the default gaussian device. The test ensures that the device is properly
    applying gaussian operations and calculating the correct observables."""
    def setUp(self):
        self.dev = DefaultGaussian(wires=2, shots=0, hbar=hbar)

    def test_operation_map(self):
        """Test that default Gaussian device supports all PennyLane Gaussian CV gates."""
        self.logTestName()

        non_supported = {'FockDensityMatrix',
                         'FockStateVector',
                         'FockState',
                         'CrossKerr',
                         'CatState',
                         'CubicPhase',
                         'Kerr'}

        self.assertEqual(set(qml.ops.cv.__all__) - non_supported,
                         set(self.dev._operation_map))

    def test_expectation_map(self):
        """Test that default Gaussian device supports all PennyLane Gaussian continuous expectations."""
        self.logTestName()
        self.assertEqual(set(qml.expval.cv.__all__)|{'Identity'}-{'Heterodyne'},
                         set(self.dev._expectation_map))

    def test_apply(self):
        """Test the application of gates to a state"""
        self.logTestName()

        # loop through all supported operations
        for gate_name, fn in self.dev._operation_map.items():
            log.debug("\tTesting %s gate...", gate_name)
            self.dev.reset()

            # start in the displaced squeezed state
            alpha = 0.542+0.123j
            a = abs(alpha)
            phi_a = np.angle(alpha)
            r = 0.652
            phi_r = -0.124

            self.dev.apply('DisplacedSqueezedState', wires=[0], par=[a, phi_a, r, phi_r])
            self.dev.apply('DisplacedSqueezedState', wires=[1], par=[a, phi_a, r, phi_r])

            # get the equivalent pennylane operation class
            op = qml.ops.__getattribute__(gate_name)
            # the list of wires to apply the operation to
            w = list(range(op.num_wires))

            if op.par_domain == 'A':
                # the parameter is an array
                if gate_name == 'GaussianState':
                    p = [np.array([0.432, 0.123, 0.342, 0.123]), np.diag([0.5234]*4)]
                    w = list(range(2))
                    expected_out = p
                elif gate_name == 'Interferometer':
                    w = list(range(2))
                    p = [U]
                    S = fn(*p)
                    expected_out = S @ self.dev._state[0], S @ self.dev._state[1] @ S.T
            else:
                # the parameter is a float
                p = [0.432423, -0.12312, 0.324, 0.751][:op.num_params]

                if gate_name == 'Displacement':
                    alpha = p[0]*np.exp(1j*p[1])
                    state = self.dev._state
                    mu = state[0].copy()
                    mu[w[0]] += alpha.real*np.sqrt(2*hbar)
                    mu[w[0]+2] += alpha.imag*np.sqrt(2*hbar)
                    expected_out = mu, state[1]
                elif 'State' in gate_name:
                    mu, cov = fn(*p, hbar=hbar)
                    expected_out = self.dev._state
                    expected_out[0][[w[0], w[0]+2]] = mu

                    ind = np.concatenate([np.array([w[0]]), np.array([w[0]])+2])
                    rows = ind.reshape(-1, 1)
                    cols = ind.reshape(1, -1)
                    expected_out[1][rows, cols] = cov
                else:
                    # if the default.gaussian is an operation accepting parameters,
                    # initialise it using the parameters generated above.
                    S = fn(*p)

                    # calculate the expected output
                    if op.num_wires == 1:
                        # reorder from symmetric ordering to xp-ordering
                        S = block_diag(S, np.identity(2))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]

                    expected_out = S @ self.dev._state[0], S @ self.dev._state[1] @ S.T

            self.dev.apply(gate_name, wires=w, par=p)

            # verify the device is now in the expected state
            self.assertAllAlmostEqual(self.dev._state[0], expected_out[0], delta=self.tol)
            self.assertAllAlmostEqual(self.dev._state[1], expected_out[1], delta=self.tol)

    def test_apply_errors(self):
        """Test that apply fails for incorrect state preparation"""
        self.logTestName()

        with self.assertRaisesRegex(ValueError, 'incorrect size for the number of subsystems'):
            p = [thermal_state(0.5)]
            self.dev.apply('GaussianState', wires=[0], par=[p])

        with self.assertRaisesRegex(ValueError, 'incorrect number of subsystems'):
            p = U
            self.dev.apply('Interferometer', wires=[0], par=[p])

        with self.assertRaisesRegex(ValueError, 'Only 2-mode interferometers are currently supported.'):
            p = U2
            dev = DefaultGaussian(wires=4, shots=0, hbar=hbar)
            self.dev.apply('Interferometer', wires=[0, 1, 2, 3], par=[p])

    def test_expectation(self):
        """Test that expectation values are calculated correctly"""
        self.logTestName()

        dev = qml.device('default.gaussian', wires=1, hbar=hbar)

        # test correct mean and variance for <n> of a displaced thermal state
        nbar = 0.5431
        alpha = 0.324-0.59j
        dev.apply('ThermalState', wires=[0], par=[nbar])
        dev.apply('Displacement', wires=[0], par=[alpha, 0])
        mean = dev.expval('MeanPhoton', [0], [])
        self.assertAlmostEqual(mean, np.abs(alpha)**2+nbar, delta=self.tol)
        # self.assertAlmostEqual(var, nbar**2+nbar+np.abs(alpha)**2*(1+2*nbar), delta=self.tol)

        # test correct mean and variance for Homodyne P measurement
        alpha = 0.324-0.59j
        dev.apply('CoherentState', wires=[0], par=[alpha])
        mean = dev.expval('P', [0], [])
        self.assertAlmostEqual(mean, alpha.imag*np.sqrt(2*hbar), delta=self.tol)
        # self.assertAlmostEqual(var, hbar/2, delta=self.tol)

        # test correct mean and variance for Homodyne measurement
        mean = dev.expval('Homodyne', [0], [np.pi/2])
        self.assertAlmostEqual(mean, alpha.imag*np.sqrt(2*hbar), delta=self.tol)
        # self.assertAlmostEqual(var, hbar/2, delta=self.tol)

        # test correct mean and variance for number state expectation |<n|alpha>|^2
        # on a coherent state
        for n in range(3):
            mean = dev.expval('NumberState', [0], [np.array([n])])
            expected = np.abs(np.exp(-np.abs(alpha)**2/2)*alpha**n/np.sqrt(fac(n)))**2
            self.assertAlmostEqual(mean, expected, delta=self.tol)

        # test correct mean and variance for number state expectation |<n|S(r)>|^2
        # on a squeezed state
        n = 1
        r = 0.4523
        dev.apply('SqueezedState', wires=[0], par=[r, 0])
        mean = dev.expval('NumberState', [0], [np.array([2*n])])
        expected = np.abs(np.sqrt(fac(2*n))/(2**n*fac(n))*(-np.tanh(r))**n/np.sqrt(np.cosh(r)))**2
        self.assertAlmostEqual(mean, expected, delta=self.tol)

    def test_reduced_state(self):
        """Test reduced state"""
        self.logTestName()

        # Test error is raised if requesting a non-existant subsystem
        with self.assertRaisesRegex(ValueError, "specified wires cannot be larger than the number of subsystems"):
            self.dev.reduced_state([6, 4])

        # Test requesting via an integer
        res = self.dev.reduced_state(0)
        expected = self.dev.reduced_state([0])
        self.assertAllAlmostEqual(res[0], expected[0], delta=self.tol)
        self.assertAllAlmostEqual(res[1], expected[1], delta=self.tol)

        # Test requesting all wires returns the full state
        res = self.dev.reduced_state([0, 1])
        expected = self.dev._state
        self.assertAllAlmostEqual(res[0], expected[0], delta=self.tol)
        self.assertAllAlmostEqual(res[1], expected[1], delta=self.tol)


class TestDefaultGaussianIntegration(BaseTest):
    """Integration tests for default.gaussian. This test ensures it integrates
    properly with the PennyLane interface, in particular QNode."""

    def test_load_default_gaussian_device(self):
        """Test that the default plugin loads correctly"""
        self.logTestName()

        dev = qml.device('default.gaussian', wires=2, hbar=2)
        self.assertEqual(dev.num_wires, 2)
        self.assertEqual(dev.shots, 0)
        self.assertEqual(dev.hbar, 2)
        self.assertEqual(dev.short_name, 'default.gaussian')

    def test_args(self):
        """Test that the plugin requires correct arguments"""
        self.logTestName()

        with self.assertRaisesRegex(TypeError, "missing 1 required positional argument: 'wires'"):
            qml.device('default.gaussian')

    def test_unsupported_gates(self):
        """Test error is raised with unsupported gates"""
        self.logTestName()
        dev = qml.device('default.gaussian', wires=2)

        gates = set(dev._operation_map.keys())
        all_gates = {m[0] for m in inspect.getmembers(qml.ops, inspect.isclass)}

        for g in all_gates - gates:
            op = getattr(qml.ops, g)

            if op.num_wires == 0:
                wires = [0]
            else:
                wires = list(range(op.num_wires))

            @qml.qnode(dev)
            def circuit(*x):
                """Test quantum function"""
                x = prep_par(x, op)
                op(*x, wires=wires)

                if issubclass(op, qml.operation.CV):
                    return qml.expval.X(0)

                return qml.expval.PauliZ(0)

            with self.assertRaisesRegex(qml.DeviceError, "Gate {} not supported on device default.gaussian".format(g)):
                x = np.random.random([op.num_params])
                circuit(*x)

    def test_unsupported_observables(self):
        """Test error is raised with unsupported observables"""
        self.logTestName()
        dev = qml.device('default.gaussian', wires=2)

        obs = set(dev._expectation_map.keys())
        all_obs = set(qml.expval.__all__)

        for g in all_obs - obs:
            op = getattr(qml.expval, g)

            if op.num_wires == 0:
                wires = [0]
            else:
                wires = list(range(op.num_wires))

            @qml.qnode(dev)
            def circuit(*x):
                """Test quantum function"""
                x = prep_par(x, op)
                return op(*x, wires=wires)

            with self.assertRaisesRegex(qml.DeviceError, "Expectation {} not supported on device default.gaussian".format(g)):
                x = np.random.random([op.num_params])
                circuit(*x)

    def test_gaussian_circuit(self):
        """Test that the default gaussian plugin provides correct result for simple circuit"""
        self.logTestName()
        dev = qml.device('default.gaussian', wires=1)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.Displacement(x, 0, wires=0)
            return qml.expval.X(0)

        self.assertAlmostEqual(circuit(p), p*np.sqrt(2*hbar), delta=self.tol)

    def test_gaussian_identity(self):
        """Test that the default gaussian plugin provides correct result for the identity expectation"""
        self.logTestName()
        dev = qml.device('default.gaussian', wires=1)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.Displacement(x, 0, wires=0)
            return qml.expval.Identity(0)

        self.assertAlmostEqual(circuit(p), 1, delta=self.tol)

    def test_nonzero_shots(self):
        """Test that the default gaussian plugin provides correct result for high shot number"""
        self.logTestName()

        shots = 10**4
        dev = qml.device('default.gaussian', wires=1, shots=shots)

        p = 0.543

        @qml.qnode(dev)
        def circuit(x):
            """Test quantum function"""
            qml.Displacement(x, 0, wires=0)
            return qml.expval.X(0)

        runs = []
        for _ in range(100):
            runs.append(circuit(p))

        self.assertAlmostEqual(np.mean(runs), p*np.sqrt(2*hbar), delta=0.01)

    def test_supported_gates(self):
        """Test that all supported gates work correctly"""
        self.logTestName()
        a = 0.312

        dev = qml.device('default.gaussian', wires=2)

        for g, qop in dev._operation_map.items():
            log.debug('\tTesting gate %s...', g)
            self.assertTrue(dev.supported(g))
            dev.reset()

            op = getattr(qml.ops, g)
            if op.num_wires == 0:
                wires = list(range(2))
            else:
                wires = list(range(op.num_wires))

            @qml.qnode(dev)
            def circuit(*x):
                """Reference quantum function"""
                qml.Displacement(a, 0, wires=[0])
                op(*x, wires=wires)
                return qml.expval.X(0)

            # compare to reference result
            def reference(*x):
                """reference circuit"""
                if g == 'GaussianState':
                    return x[0][0]

                if g == 'Displacement':
                    alpha = x[0]*np.exp(1j*x[1])
                    return (alpha+a).real*np.sqrt(2*hbar)

                if 'State' in g:
                    mu, _ = qop(*x, hbar=hbar)
                    return mu[0]

                S = qop(*x)

                # calculate the expected output
                if op.num_wires == 1:
                    S = block_diag(S, np.identity(2))[:, [0, 2, 1, 3]][[0, 2, 1, 3]]

                return (S @ np.array([a.real, a.imag, 0, 0])*np.sqrt(2*hbar))[0]

            if g == 'GaussianState':
                p = [np.array([0.432, 0.123, 0.342, 0.123]), np.diag([0.5234]*4)]
            elif g == 'Interferometer':
                p = [np.array(U)]
            else:
                p = [0.432423, -0.12312, 0.324, 0.763][:op.num_params]

            self.assertAllEqual(circuit(*p), reference(*p))


if __name__ == '__main__':
    print('Testing PennyLane version ' + qml.version() + ', default.gaussian plugin.')
    # run the tests in this file
    suite = unittest.TestSuite()
    for t in (TestAuxillaryFunctions,
              TestGates,
              TestStates,
              TestDefaultGaussianDevice,
              TestDefaultGaussianIntegration):
        ttt = unittest.TestLoader().loadTestsFromTestCase(t)
        suite.addTests(ttt)
    unittest.TextTestRunner().run(suite)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.plugin.DefaultGaussian` device.
"""
# pylint: disable=protected-access,cell-var-from-loop
import unittest
import inspect
import logging as log

from pennylane import numpy as np
from scipy.linalg import block_diag

from defaults import pennylane as qml, BaseTest
from pennylane.ops import cv


log.getLogger('defaults')

hbar = 2
phis = np.linspace(-2 * np.pi, 2 * np.pi, 11)
mags = np.linspace(0., 1., 7)
s_vals = np.linspace(-3,3,13)

nongaussian_gates = [cv.Kerr, cv.CrossKerr, cv.CubicPhase]

class TestHeisenberg(BaseTest):
    """Tests for the Heisenberg representation of gates."""

    def test_rotation_heisenberg(self):
        """Tests the Heisenberg representation of the Rotation gate."""
        self.logTestName()

        for phi in phis:
            matrix = cv.Rotation._heisenberg_rep([phi])
            true_matrix = np.array([[1, 0, 0],
                                    [0, np.cos(phi), -np.sin(phi)],
                                    [0, np.sin(phi), np.cos(phi)]])
            self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)


    def test_squeezing_heisenberg(self):
        """Tests the Heisenberg representation of the Squeezing gate."""
        self.logTestName()

        for r in mags:
            for phi in phis:
                matrix = cv.Squeezing._heisenberg_rep([r,phi])
                true_matrix = np.array([[1, 0, 0],
                                        [0, np.cosh(r) - np.cos(phi) * np.sinh(r), - np.sin(phi) * np.sinh(r)],
                                        [0, -np.sin(phi) * np.sinh(r), np.cosh(r) + np.cos(phi) * np.sinh(r)]])
                self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)


    def test_displacement_heisenberg(self):
        """Tests the Heisenberg representation of the Displacement gate."""
        self.logTestName()

        for r in mags:
            for phi in phis:
                matrix = cv.Displacement._heisenberg_rep([r,phi])
                true_matrix = np.array([[1, 0, 0],
                                        [np.sqrt(2 * hbar) * r * np.cos(phi), 1, 0],
                                        [np.sqrt(2 * hbar) * r * np.sin(phi), 0, 1]])
                self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)


    def test_beamsplitter_heisenberg(self):
        """Tests the Heisenberg representation of the Beamsplitter gate."""
        self.logTestName()

        for theta in phis:
            for phi in phis:
                matrix = cv.Beamsplitter._heisenberg_rep([theta,phi])
                true_matrix = np.array([[1, 0, 0, 0, 0],
                                        [0, np.cos(theta), 0, -np.cos(phi) * np.sin(theta), -np.sin(phi) * np.sin(theta)],
                                        [0, 0, np.cos(theta), np.sin(phi) * np.sin(theta), -np.cos(phi) * np.sin(theta)],
                                        [0, np.cos(phi) * np.sin(theta), -np.sin(phi) * np.sin(theta), np.cos(theta), 0],
                                        [0, np.sin(phi) * np.sin(theta), np.cos(phi) * np.sin(theta), 0, np.cos(theta)]])
                self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)


    def test_two_mode_squeezing_heisenberg(self):
        """Tests the Heisenberg representation of the Beamsplitter gate."""
        self.logTestName()

        for r in mags:
            for phi in phis:
                matrix = cv.TwoModeSqueezing._heisenberg_rep([r,phi])
                true_matrix = np.array([[1, 0, 0, 0, 0],
                                        [0, np.cosh(r), 0, np.cos(phi) * np.sinh(r), np.sin(phi) * np.sinh(r)],
                                        [0, 0, np.cosh(r), np.sin(phi) * np.sinh(r), -np.cos(phi) * np.sinh(r)],
                                        [0, np.cos(phi) * np.sinh(r), np.sin(phi) * np.sinh(r), np.cosh(r), 0],
                                        [0, np.sin(phi) * np.sinh(r), -np.cos(phi) * np.sinh(r), 0, np.cosh(r)]])
                self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)


    def test_quadratic_phase_heisenberg(self):
        """Tests the Heisenberg representation of the QuadraticPhase gate."""
        self.logTestName()

        for s in s_vals:
            matrix = cv.QuadraticPhase._heisenberg_rep([s])
            true_matrix = np.array([[1, 0, 0],
                                    [0, 1, 0],
                                    [0, s, 1]])
            self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)


    def test_controlled_addition_heisenberg(self):
        """Tests the Heisenberg representation of the ControlledAddition gate."""
        self.logTestName()

        for s in s_vals:
            matrix = cv.ControlledAddition._heisenberg_rep([s])
            true_matrix = np.array([[1, 0, 0, 0, 0],
                                    [0, 1, 0, 0, 0],
                                    [0, 0, 1, 0, -s],
                                    [0, s, 0, 1, 0],
                                    [0, 0, 0, 0, 1]])
            self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)


    def test_controlled_phase_heisenberg(self):
        """Tests the Heisenberg representation of the ControlledPhase gate."""
        self.logTestName()

        for s in s_vals:
            matrix = cv.ControlledPhase._heisenberg_rep([s])
            true_matrix = np.array([[1, 0, 0, 0, 0],
                                    [0, 1, 0, 0, 0],
                                    [0, 0, 1, s, 0],
                                    [0, 0, 0, 1, 0],
                                    [0, s, 0, 0, 1]])
            self.assertAllAlmostEqual(matrix, true_matrix, delta=self.tol)

class TestNonGaussian(BaseTest):
    """Tests that non-Gaussian gates are properly handled."""

    def test_heisenberg_rep_nonguassian(self):
        """Tests that the `_heisenberg_rep` for a non-Gaussian gates is None"""
        for op in nongaussian_gates:
            self.assertTrue(op._heisenberg_rep(*[0.1] * op.num_params) is None)

    def test_heisenberg_transformation_nongaussian(self):
        """Tests that proper exceptions are raised if we try to call the Heisenberg
        transformation of non-Gaussian gates."""
        self.logTestName()

        with self.assertRaisesRegex(RuntimeError, 'is not a Gaussian operation'):
            for op in nongaussian_gates:
                op_ = op(*[0.1] * op.num_params, [0] * op.num_wires, do_queue=False)
                op_.heisenberg_tr(op.num_wires)

if __name__ == '__main__':
    print('Testing PennyLane version ' + qml.version() + ', default.gaussian plugin.')
    # run the tests in this file
    suite = unittest.TestSuite()
    for t in (TestHeisenberg,TestNonGaussian):
        ttt = unittest.TestLoader().loadTestsFromTestCase(t)
        suite.addTests(ttt)
    unittest.TextTestRunner().run(suite)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Unit tests for the :mod:`pennylane.template` module.
"""
# pylint: disable=protected-access,cell-var-from-loop
import unittest
from unittest.mock import patch
import logging as log
import itertools as it

from pennylane import numpy as np
from pennylane.qnode import QuantumFunctionError
from pennylane.plugins import DefaultGaussian
from defaults import pennylane as qml, BaseTest

log.getLogger('defaults')


class DummyDevice(DefaultGaussian):
    """Dummy device to allow Kerr operations"""
    _operation_map = DefaultGaussian._operation_map.copy()
    _operation_map['Kerr'] = lambda *x, **y: np.identity(2)


class TestInterferometer(BaseTest):
    """Tests for the Interferometer from the pennylane.template module."""

    def test_exceptions(self):
        """test that exceptions are correctly raised"""
        dev = qml.device('default.gaussian', wires=1)
        varphi = [0.42342]

        @qml.qnode(dev)
        def circuit(varphi, mesh):
            qml.template.Interferometer(theta=None, phi=None, varphi=varphi, mesh=mesh, wires=0)
            return qml.expval.MeanPhoton(0)

        with self.assertRaisesRegex(QuantumFunctionError, "The mesh parameter influences the "
        "circuit architecture and can not be passed as a QNode parameter."):
            circuit(varphi, 'rectangular')

        @qml.qnode(dev)
        def circuit(varphi, bs):
            qml.template.Interferometer(theta=None, phi=None, varphi=varphi, beamsplitter=bs, wires=0)
            return qml.expval.MeanPhoton(0)

        with self.assertRaisesRegex(QuantumFunctionError, "The beamsplitter parameter influences the "
        "circuit architecture and can not be passed as a QNode parameter."):
            circuit(varphi, 'clements')

    def test_clements_beamsplitter_convention(self):
        """test the beamsplitter convention"""
        N = 2
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        theta = [0.321]
        phi = [0.234]
        varphi = [0.42342]

        def circuit_rect(varphi):
            qml.template.Interferometer(theta, phi, varphi, mesh='rectangular', beamsplitter='clements', wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        def circuit_tria(varphi):
            qml.template.Interferometer(theta, phi, varphi, mesh='triangular', beamsplitter='clements', wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        for c in [circuit_rect, circuit_tria]:
            qnode = qml.QNode(c, dev)
            self.assertAllAlmostEqual(qnode(varphi), [0, 0], delta=self.tol)

            queue = qnode.queue
            self.assertEqual(len(queue), 3)

            self.assertTrue(isinstance(qnode.queue[0], qml.Rotation))
            self.assertAllEqual(qnode.queue[0].parameters, phi)

            self.assertTrue(isinstance(qnode.queue[1], qml.Beamsplitter))
            self.assertAllEqual(qnode.queue[1].parameters, [theta[0], 0])

            self.assertTrue(isinstance(qnode.queue[2], qml.Rotation))
            self.assertAllEqual(qnode.queue[2].parameters, varphi)

    def test_one_mode(self):
        """Test that a one mode interferometer correctly gives a rotation gate"""
        dev = qml.device('default.gaussian', wires=1)
        varphi = [0.42342]

        def circuit(varphi):
            qml.template.Interferometer(theta=None, phi=None, varphi=varphi, wires=0)
            return qml.expval.MeanPhoton(0)

        qnode = qml.QNode(circuit, dev)
        self.assertAllAlmostEqual(qnode(varphi), 0, delta=self.tol)

        queue = qnode.queue
        self.assertEqual(len(queue), 1)
        self.assertTrue(isinstance(qnode.queue[0], qml.Rotation))
        self.assertAllEqual(qnode.queue[0].parameters, varphi)

    def test_two_mode_rect(self):
        """Test that a two mode interferometer using the rectangular mesh
        correctly gives a beamsplitter+rotation gate"""
        N = 2
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        theta = [0.321]
        phi = [0.234]
        varphi = [0.42342]

        def circuit(varphi):
            qml.template.Interferometer(theta, phi, varphi, wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        qnode = qml.QNode(circuit, dev)
        self.assertAllAlmostEqual(qnode(varphi), [0, 0], delta=self.tol)

        queue = qnode.queue
        self.assertEqual(len(queue), 2)

        self.assertTrue(isinstance(qnode.queue[0], qml.Beamsplitter))
        self.assertAllEqual(qnode.queue[0].parameters, theta+phi)

        self.assertTrue(isinstance(qnode.queue[1], qml.Rotation))
        self.assertAllEqual(qnode.queue[1].parameters, varphi)

    def test_two_mode_triangular(self):
        """Test that a two mode interferometer using the triangular mesh
        correctly gives a beamsplitter+rotation gate"""
        N = 2
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        theta = [0.321]
        phi = [0.234]
        varphi = [0.42342]

        def circuit(varphi):
            qml.template.Interferometer(theta, phi, varphi, mesh='triangular', wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        qnode = qml.QNode(circuit, dev)
        self.assertAllAlmostEqual(qnode(varphi), [0, 0], delta=self.tol)

        queue = qnode.queue
        self.assertEqual(len(queue), 2)

        self.assertTrue(isinstance(qnode.queue[0], qml.Beamsplitter))
        self.assertAllEqual(qnode.queue[0].parameters, theta+phi)

        self.assertTrue(isinstance(qnode.queue[1], qml.Rotation))
        self.assertAllEqual(qnode.queue[1].parameters, varphi)

    def test_two_mode_rect_overparameterised(self):
        """Test that a two mode interferometer using the rectangular mesh
        correctly gives a beamsplitter+2 rotation gates when requested"""
        N = 2
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        theta = [0.321]
        phi = [0.234]
        varphi = [0.42342, 0.543]

        def circuit(varphi):
            qml.template.Interferometer(theta, phi, varphi, wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        qnode = qml.QNode(circuit, dev)
        self.assertAllAlmostEqual(qnode(varphi), [0, 0], delta=self.tol)

        queue = qnode.queue
        self.assertEqual(len(queue), 3)

        self.assertTrue(isinstance(qnode.queue[0], qml.Beamsplitter))
        self.assertAllEqual(qnode.queue[0].parameters, theta+phi)

        self.assertTrue(isinstance(qnode.queue[1], qml.Rotation))
        self.assertAllEqual(qnode.queue[1].parameters, [varphi[0]])

        self.assertTrue(isinstance(qnode.queue[2], qml.Rotation))
        self.assertAllEqual(qnode.queue[2].parameters, [varphi[1]])

    def test_three_mode(self):
        """Test that a three mode interferometer using either mesh gives the correct gates"""
        N = 3
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        theta = [0.321, 0.4523, 0.21321]
        phi = [0.234, 0.324, 0.234]
        varphi = [0.42342, 0.234]

        def circuit_rect(varphi):
            qml.template.Interferometer(theta, phi, varphi, wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        def circuit_tria(varphi):
            qml.template.Interferometer(theta, phi, varphi, wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        for c in [circuit_rect, circuit_tria]:
            # test both meshes (both give identical results for the 3 mode case).
            qnode = qml.QNode(c, dev)
            self.assertAllAlmostEqual(qnode(varphi), [0]*N, delta=self.tol)

            queue = qnode.queue
            self.assertEqual(len(queue), 5)

            expected_bs_wires = [[0, 1], [1, 2], [0, 1]]

            for idx, op in enumerate(qnode.queue[:3]):
                self.assertTrue(isinstance(op, qml.Beamsplitter))
                self.assertAllEqual(op.parameters, [theta[idx], phi[idx]])
                self.assertAllEqual(op.wires, expected_bs_wires[idx])

            for idx, op in enumerate(qnode.queue[3:]):
                self.assertTrue(isinstance(op, qml.Rotation))
                self.assertAllEqual(op.parameters, [varphi[idx]])
                self.assertAllEqual(op.wires, [idx])

    def test_four_mode_rect(self):
        """Test that a 4 mode interferometer using rectangular mesh gives the correct gates"""
        N = 4
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        theta = [0.321, 0.4523, 0.21321, 0.123, 0.5234, 1.23]
        phi = [0.234, 0.324, 0.234, 1.453, 1.42341, -0.534]
        varphi = [0.42342, 0.234, 0.4523]

        def circuit_rect(varphi):
            qml.template.Interferometer(theta, phi, varphi, wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        qnode = qml.QNode(circuit_rect, dev)
        self.assertAllAlmostEqual(qnode(varphi), [0]*N, delta=self.tol)

        queue = qnode.queue
        self.assertEqual(len(queue), 9)

        expected_bs_wires = [[0, 1], [2, 3], [1, 2], [0, 1], [2, 3], [1, 2]]

        for idx, op in enumerate(qnode.queue[:6]):
            self.assertTrue(isinstance(op, qml.Beamsplitter))
            self.assertAllEqual(op.parameters, [theta[idx], phi[idx]])
            self.assertAllEqual(op.wires, expected_bs_wires[idx])

        for idx, op in enumerate(qnode.queue[6:]):
            self.assertTrue(isinstance(op, qml.Rotation))
            self.assertAllEqual(op.parameters, [varphi[idx]])
            self.assertAllEqual(op.wires, [idx])

    def test_four_mode_triangular(self):
        """Test that a 4 mode interferometer using triangular mesh gives the correct gates"""
        N = 4
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        theta = [0.321, 0.4523, 0.21321, 0.123, 0.5234, 1.23]
        phi = [0.234, 0.324, 0.234, 1.453, 1.42341, -0.534]
        varphi = [0.42342, 0.234, 0.4523]

        def circuit_tria(varphi):
            qml.template.Interferometer(theta, phi, varphi, mesh='triangular', wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        qnode = qml.QNode(circuit_tria, dev)
        self.assertAllAlmostEqual(qnode(varphi), [0]*N, delta=self.tol)

        queue = qnode.queue
        self.assertEqual(len(queue), 9)

        expected_bs_wires = [[2, 3], [1, 2], [0, 1], [2, 3], [1, 2], [2, 3]]

        for idx, op in enumerate(qnode.queue[:6]):
            self.assertTrue(isinstance(op, qml.Beamsplitter))
            self.assertAllEqual(op.parameters, [theta[idx], phi[idx]])
            self.assertAllEqual(op.wires, expected_bs_wires[idx])

        for idx, op in enumerate(qnode.queue[6:]):
            self.assertTrue(isinstance(op, qml.Rotation))
            self.assertAllEqual(op.parameters, [varphi[idx]])
            self.assertAllEqual(op.wires, [idx])

    def test_integration(self):
        """test integration with PennyLane and gradient calculations"""
        N = 4
        wires = range(N)
        dev = qml.device('default.gaussian', wires=N)

        sq = np.array([[0.8734294 , 0.96854066],
                       [0.86919454, 0.53085569],
                       [0.23272833, 0.0113988 ],
                       [0.43046882, 0.40235136]])

        theta = np.array([3.28406182, 3.0058243, 3.48940764, 3.41419504, 4.7808479, 4.47598146])
        phi = np.array([3.89357744, 2.67721355, 1.81631197, 6.11891294, 2.09716418, 1.37476761])
        varphi = np.array([0.4134863, 6.17555778, 0.80334114, 2.02400747])

        @qml.qnode(dev)
        def circuit(theta, phi, varphi):
            for w in wires:
                qml.Squeezing(sq[w][0], sq[w][1], wires=w)

            qml.template.Interferometer(theta, phi, varphi, wires=wires)
            return [qml.expval.MeanPhoton(w) for w in wires]

        res = circuit(theta, phi, varphi)
        expected = np.array([0.96852694, 0.23878521, 0.82310606, 0.16547786])
        self.assertAllAlmostEqual(res, expected, delta=self.tol)

        res = qml.jacobian(circuit, 0)(theta, phi, varphi)
        expected = np.array([[-6.18547621e-03, -3.20488416e-04, -4.20274088e-02, -6.21819677e-02, 0.00000000e+00, 0.00000000e+00],
                            [3.55439469e-04, 3.89820231e-02, -3.35281297e-03, 7.93009278e-04, 8.30347900e-02, -3.45150712e-01],
                            [5.44893709e-03, 9.30878023e-03, -5.33374090e-01, 6.13889584e-02, -1.16931386e-01, 3.45150712e-01],
                            [3.81099656e-04, -4.79703149e-02, 5.78754312e-01, 0.00000000e+00, 3.38965962e-02, 0.00000000e+00]])
        self.assertAllAlmostEqual(res, expected, delta=self.tol)


class TestCVNeuralNet(BaseTest):
    """Tests for the CVNeuralNet from the pennylane.template module."""
    num_subsystems = 4

    def setUp(self):
        super().setUp()
        self.N = 4
        self.depth = 2
        self.weights = [
            [
                np.array([ 5.48791879, 6.08552046, 5.46131036, 3.33546468, 1.46227521, 0.0716208 ]),
                np.array([ 2.70471535, 2.52804815, 3.28406182, 3.0058243 , 3.48940764, 3.41419504]),
                np.array([ 4.7808479 , 4.47598146, 3.89357744, 2.67721355]),
                np.array([ 0.27344502, 0.68431314, 0.30026443, 0.23128064]),
                np.array([ 0.4134863 , 6.17555778, 0.80334114, 2.02400747, 0.44574704, 1.41227118]),
                np.array([ 2.47328111, 5.63064513, 2.17059932, 6.1873632 , 0.18052879, 2.20970037]),
                np.array([ 2.3936353 , 4.80135971, 5.89867895, 2.00867023, 2.71732643, 1.69737575]),
                np.array([ 5.03318258, 4.01017269, 0.43159284, 3.7928101 ]),
                np.array([ 1.61159166, 0.1608155 , 0.96535086, 1.60132783]),
                np.array([ 6.21267547, 3.71076099, 0.34060195, 2.86031556]),
                np.array([ 0.1376345 , 0.22541113, 0.14306356, 0.13019402])
            ],
            [
                np.array([ 3.36869403, 0.63074883, 4.59400392, 5.9040016 , 5.92704296, 2.35455147]),
                np.array([ 3.74320919, 4.15936005, 3.20807161, 2.95870535, 0.05574621, 0.42660569]),
                np.array([ 2.73203094, 2.71115444, 1.16794164, 3.32823666]),
                np.array([ 0.45945175, 0.53255468, 0.28383751, 0.34263728]),
                np.array([ 5.16969442, 3.6890488 , 4.43916808, 3.20808287, 5.21543123, 4.52815349]),
                np.array([ 5.44288268, 1.27806129, 1.87574979, 2.98956484, 3.10140853, 3.81814174]),
                np.array([ 5.14552399, 3.31578667, 5.90119363, 4.54515204, 1.12316345, 3.89384963]),
                np.array([ 3.5329307 , 4.79661266, 5.0683084 , 1.87631749]),
                np.array([ 0.36293094, 1.30725604, 0.11578591, 1.5983082 ]),
                np.array([ 3.20443756, 6.26536946, 6.18450567, 1.50406923]),
                np.array([ 0.26999146, 0.26256351, 0.14722687, 0.23137066])]
            ]

    def test_CVNeuralNet_integration(self):
        """integration test for the CVNeuralNet."""
        dev = DummyDevice(wires=self.num_subsystems)

        def circuit(weights):
            qml.template.CVNeuralNet(weights, wires=range(self.N))
            return qml.expval.X(wires=0)

        qnode = qml.QNode(circuit, dev)

        # execution test
        qnode(self.weights)
        queue = qnode.queue

        # Test that gates appear in the right order for each layer:
        # BS-R-S-BS-R-D-K
        for l in range(self.depth):
            gates = [qml.Beamsplitter, qml.Rotation, qml.Squeezing,
                     qml.Beamsplitter, qml.Rotation, qml.Displacement]

            # count the position of each group of gates in the layer
            num_gates_per_type = [0, 6, 4, 4, 6, 4, 4, 4]
            s = np.cumsum(num_gates_per_type)
            gc = l*sum(num_gates_per_type)+np.array(list(zip(s[:-1], s[1:])))

            # loop through expected gates
            for idx, g in enumerate(gates):
                # loop through where these gates should be in the queue
                for opidx, op in enumerate(queue[gc[idx, 0]:gc[idx, 1]]):
                    # check that op in queue is correct gate
                    self.assertTrue(isinstance(op, g))

                    # test that the parameters are correct
                    res_params = op.parameters

                    if idx == 0:
                        # first BS
                        exp_params = [self.weights[l][0][opidx], self.weights[l][1][opidx]]
                    elif idx == 1:
                        # first rot
                        exp_params = [self.weights[l][2][opidx]]
                    elif idx == 2:
                        # squeezing
                        exp_params = [self.weights[l][3][opidx], self.weights[l][4][opidx]]
                    elif idx == 3:
                        # second BS
                        exp_params = [self.weights[l][5][opidx], self.weights[l][6][opidx]]
                    elif idx == 4:
                        # second rot
                        exp_params = [self.weights[l][7][opidx]]
                    elif idx == 5:
                        # displacement
                        exp_params = [self.weights[l][8][opidx], self.weights[l][9][opidx]]

                    self.assertEqual(res_params, exp_params)


class TestStronglyEntanglingCircuit(BaseTest):
    """Tests for the StronglyEntanglingCircuit from the pennylane.template module."""

    def test_integration(self):
        """integration test for the StronglyEntanglingCircuit."""
        np.random.seed(12)
        num_layers = 2

        for num_wires in range(2, 4):
            dev = qml.device('default.qubit', wires=num_wires)
            weights = np.random.randn(num_layers, num_wires, 3)
            def circuit(weights):
                qml.template.StronglyEntanglingCircuit(weights, True, wires=range(num_wires))
                return qml.expval.PauliZ(0)

            qnode = qml.QNode(circuit, dev)
            qnode(weights)
            queue = qnode.queue

            # Test that gates appear in the right order
            exp_gates = [qml.Rot]*num_wires + [qml.CNOT]*num_wires
            exp_gates *= num_layers
            res_gates = [op for op in queue]

            for op1, op2 in zip(res_gates, exp_gates):
                self.assertTrue(isinstance(op1, op2))

            # test the device parameters
            for l in range(num_layers):
                layer_ops = queue[2*l*num_wires:2*(l+1)*num_wires]

                # check each rotation gate parameter
                for n in range(num_wires):
                    res_params = layer_ops[n].parameters
                    exp_params = weights[l, n, :]
                    self.assertAllEqual(res_params, exp_params)

    def test_execution(self):
        """Tests the StronglyEntanglingCircuit for various parameters."""
        np.random.seed(0)
        outcomes = []

        for num_wires in range(2, 4):
            for num_layers in range(1, 3):

                dev = qml.device('default.qubit', wires=num_wires)
                weights = np.random.randn(num_layers, num_wires, 3)

                @qml.qnode(dev)
                def circuit(weights, x=None):
                    qml.BasisState(x, wires=range(num_wires))
                    qml.template.StronglyEntanglingCircuit(weights, True, wires=range(num_wires))
                    return qml.expval.PauliZ(0)

                outcomes.append(circuit(weights, x=np.array(np.random.randint(0, 1, num_wires))))

        res = np.array(outcomes)
        expected = np.array([-0.29242496, 0.22129055, 0.07540091, -0.77626557])
        self.assertAllAlmostEqual(res, expected, delta=self.tol)


if __name__ == '__main__':
    print('Testing PennyLane version ' + qml.version() + ', pennylane.template.')
    # run the tests in this file
    suite = unittest.TestSuite()
    for t in (TestInterferometer, TestCVNeuralNet, TestStronglyEntanglingCircuit):
        ttt = unittest.TestLoader().loadTestsFromTestCase(t)
        suite.addTests(ttt)
    unittest.TextTestRunner().run(suite)



# Copyright 2018 Xanadu Quantum Technologies Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Configuration
=============

**Module name:** :mod:`pennylane.configuration`

.. currentmodule:: pennylane.configuration

This module contains the :class:`Configuration` class, which is used to
load, store, save, and modify configuration options for PennyLane and all
supported plugins and devices.

Behaviour
---------

On first import, PennyLane attempts to load the configuration file `config.toml`, by
scanning the following three directories in order of preference:

1. The current directory
2. The path stored in the environment variable ``PENNYLANE_CONF``
3. The default user configuration directory:

   * On Linux: ``~/.config/pennylane``
   * On Windows: ``~C:\Users\USERNAME\AppData\Local\Xanadu\pennylane``
   * On MacOS: ``~/Library/Preferences/pennylane``

If no configuration file is found, a warning message will be displayed in the logs,
and all device parameters will need to be passed as keyword arguments when
loading the device.

The user can access the initialized configuration via `pennylane.config`, view the
loaded configuration filepath, print the configurations options, access and modify
them via keys (i.e., ``pennylane.config['main.shots']``), and save/load new configuration files.

Configuration files
-------------------

The configuration file `config.toml` uses the `TOML standard <https://github.com/toml-lang/toml>`_,
and has the following format:

.. code-block:: toml

    [main]
    # Global PennyLane options.
    # Affects every loaded plugin if applicable.
    shots = 0

    [strawberryfields.global]
    # Options for the Strawberry Fields plugin
    hbar = 1
    shots = 100

      [strawberryfields.fock]
      # Options for the Strawberry Fields Fock plugin
      cutoff_dim = 10
      hbar = 0.5

      [strawberryfields.gaussian]
      # Indentation doesn't matter in TOML files,
      # but helps provide clarity.

    [projectq.global]
    # Options for the Project Q plugin

      [projectq.simulator]
      gate_fusion = true

      [projectq.ibm]
      user = "johnsmith"
      password = "secret123"
      use_hardware = true
      device = "ibmqx4"
      num_runs = 1024

Main PennyLane options, that are passed to all loaded devices, are provided under the ``[main]``
section. Alternatively, options can be specified on a per-plugin basis, by setting the options under
``[plugin.global]``.

For example, in the above configuration file, the Strawberry Fields
devices will be loaded with a default of ``shots = 100``, rather than ``shots = 0``. Finally,
you can also specify settings on a device-by-device basis, by placing the options under the
``[plugin.device]`` settings.

Summary of methods
------------------

.. currentmodule:: pennylane.configuration.Configuration

.. autosummary::
    path
    load
    save

Helper methods
--------------

.. autosummary::
    safe_set
    safe_get

Code details
~~~~~~~~~~~~

.. currentmodule:: pennylane.configuration

"""
import os
import logging as log

import toml
from appdirs import user_config_dir

log.getLogger()


class Configuration:
    """Configuration class.

    This class is responsible for loading, saving, and storing PennyLane
    and plugin/device configurations.

    Args:
        name (str): filename of the configuration file.
        This should be a valid TOML file. You may also pass an absolute
        or a relative file path to the configuration file.
    """
    def __init__(self, name):
        # Look for an existing configuration file
        self._config = {}
        self._filepath = None
        self._name = name
        self._user_config_dir = user_config_dir('pennylane', 'Xanadu')
        self._env_config_dir = os.environ.get("PENNYLANE_CONF", "")

        # search the current directory the directory under environment
        # variable PENNYLANE_CONF, and default user config directory, in that order.
        directories = [os.curdir, self._env_config_dir, self._user_config_dir, '']
        for idx, directory in enumerate(directories):
            try:
                self._filepath = os.path.join(directory, self._name)
                self.load(self._filepath)
                break
            except FileNotFoundError:
                if idx == len(directories)-1:
                    log.warning('No PennyLane configuration file found.')

    def __str__(self):
        if self._config:
            return "{}".format(self._config)
        return ""

    def __repr__(self):
        return "PennyLane Configuration <{}>".format(self._filepath)

    @property
    def path(self):
        """Return the path of the loaded configuration file.

        Returns:
            str: If no configuration is loaded, this returns ``None``."""
        return self._filepath

    def load(self, filepath):
        """Load a configuration file.

        Args:
            filepath (str): path to the configuration file.
        """
        with open(filepath, 'r') as f:
            self._config = toml.load(f)

    def save(self, filepath):
        """Save a configuration file.

        Args:
            filepath (str): path to the configuration file.
        """
        with open(filepath, 'w') as f:
            toml.dump(self._config, f)

    def __getitem__(self, key):
        keys = key.split('.')
        return self.safe_get(self._config, *keys)

    def __setitem__(self, key, value):
        keys = key.split('.')
        self.safe_set(self._config, value, *keys)

    def __bool__(self):
        return bool(self._config)

    @staticmethod
    def safe_set(dct, value, *keys):
        """Safely set the value of a key from a nested dictionary.

        If any key provided does not exist, a dictionary containing the
        remaining keys is dynamically created and set to the required value.

        Args:
            dct (dict): the dictionary to set the value of.
            value: the value to set. Can be any valid type.
            *keys: each additional argument corresponds to a nested key.
        """
        for key in keys[:-1]:
            dct = dct.setdefault(key, {})

        dct[keys[-1]] = value

    @staticmethod
    def safe_get(dct, *keys):
        """Safely return value from a nested dictionary.

        If any key provided does not exist, an empty dictionary is returned.

        Args:
            dct (dict): the dictionary to set the value of.
            *keys: each additional argument corresponds to a nested key.

        Returns:
            value corresponding to ``dct[keys[0]][keys[1]]`` etc.
        """
        for key in keys:
            try:
                dct = dct[key]
            except KeyError:
                return {}
        return dct
